C Testing routines for the fortran version of plinda 
C Note this program serves to test both as the master 
C and a spawned process -- it differentiates itself by 
C using the arg_rdp pattern. For top level master the 
C pattern is (char*) and for the spawned worker it is 
C (char*,int).

C       Written by Tom Brown (brow7141@cs.nyu.edu) and Peter Wyckoff
C       (wyckoff@cs.nyu.edu).


	subroutine freal_main()
	implicit none
	integer g(2)

	print *,"Starting the tests"
	call fcreate_group(g,"testgids", 12, "a.f")	
	call fdestroy_group(g,12, "a.f")	
	print *,"testing gids is done"
	call testInts()
	print *,"Testing ints is done"
	call testChars()
	print *,"Testing chars is done"

	call testArgRdP()
	print *,"Testing argRdP is done"

	call testEval()
	print *,"Testing eval is done"

	call testTransactions()
	print *,"Testing transactions is done"

	call testXrecover()
	print *,"Testing xrecover is done"


	call testDoubles()
	print *,"Testing doubles is done"
	call testFloats()
	print *,"Testing floats is done"
	call testGids()
	print *,"Testing ints is done"
	call testInpRdPRd()
	print *,"Testing rd, rdp, and inp is done"
	call testConstants()
	print *,"testing constants is done"
	print *,"All tests done"
	RETURN	
	end		

	
	subroutine testConstants()
	implicit none
	integer g(2),ftuple,frint,flint
	integer frdouble,fldouble,flfloat,frfloat
	integer int
	real*8 double
	real*4 float
	integer TWOHUNDRED
	parameter(TWOHUNDRED = 200)
	real*8 ONEHUNDRED
	parameter(ONEHUNDRED = 100)
	call fcreate_group(g,"testconsts", 12, "a.f")	

	int = 0
	double = 0
	call fout(g, ftuple(1,frint(200,0)), 15, "plinda.f")
	call fin(g, ftuple(1,flint(int,0)), 16, "plinda.f")
	call fout(g, ftuple(1,frfloat(200.0,0)), 15, "plinda.f")
	call fin(g, ftuple(1,flfloat(float,0)), 16, "plinda.f")

	if(int .ne. 200 .or. float .ne. 200.0) then
	   print *,"Error **** in test consts 1"
	   print *,int
	   print *,double
	   return
	end if

	int = 0
	double = 0
	call fout(g, ftuple(1,frint(TWOHUNDRED,0)), 15, "plinda.f")
	call fin(g, ftuple(1,flint(int,0)), 16, "plinda.f")
	call fout(g, ftuple(1,frdouble(ONEHUNDRED,0)), 15, "plinda.f")
	call fin(g, ftuple(1,fldouble(double,0)), 16, "plinda.f")

	if(int .ne. 200 .or. double .ne. 100.0) then
	   print *,"Error **** in test consts 2"
	   return
	end if



	return
	end

	
	subroutine testInPRdPRd()
	implicit none
	integer int1(2),int2(2),g(2),ftuple,frint,flint
	integer*8 frdouble,frdp,finp,ret,fldouble
	real*8 double1


	call fcreate_group(g,"testinprdprd", 12, "a.f")	
	int1(1) = 1
	int1(2) = 2
	int2(1) = 3
	int2(2) = 4
	double1 = 1.1
	call fout(g, ftuple(2,frint(int1,2),frdouble(double1,0)), 15, "plinda.f")
	call frd(g, ftuple(2,flint(int2,2),frdouble(double1,0)), 16, "plinda.f")

	if(int2(1) .ne. 1 .or. int2(2) .ne. 2) then
	   print *, "Error ***** testing rd"
	   return
	end if

	int2(1) = 3
	int2(2) = 4
	ret=frdp(g, ftuple(2,flint(int2,2),frdouble(double1,0)), 16, "plinda.f")

	if(ret .ne. 1 .or. int2(1) .ne. 1 .or. int2(2) .ne. 2) then
	   print *, "Error ***** testing rdp"
	   return
	end if

	int2(1) = 3
	int2(2) = 4

	ret=finp(g, ftuple(2,flint(int2,2),frdouble(double1,0)), 16, "plinda.f")

	if(ret .ne. 1 .or. int2(1) .ne. 1 .or. int2(2) .ne. 2) then
	   print *, "Error ***** testing rdp"
	   return
	end if

	int2(1) = 3
	int2(2) = 4

	ret=finp(g, ftuple(2,flint(int2,2),frdouble(double1,0)), 16, "plinda.f")

	if(ret .ne. 0 .or. int2(1) .ne. 3 .or. int2(2) .ne. 4) then
	   print *, "Error ***** testing rdp"
	   return
	end if


	return
	end

	subroutine testXrecover()
	implicit none
	integer ftuple,flint,fxrecover
	integer x,ret
	x = 0
	ret =  fxrecover(ftuple(1,flint(x,0)),12,"plinda.f")

	if(ret .ne. 1 .or. x .ne. 2) then
	   print *,"problem with xrecover"
	end if
	return 
	end

	subroutine testTransactions()
	implicit none
	integer ftuple,frint,flint
	integer x,int1(10),int2(10),g(2)
	x = 2
	call fxstart(12,"plinda.f")
	call fxcommit(ftuple(1,frint(x,0)),12,"plinda.f")

	call fxstart(12,"plinda.f")
	call fxcommit_nostate(12,"plinda.f")
	print *,"Empty transaction was ok"
	
	call fxstart(12,"plinda.f")
	call testInts()
c	note since all in one transaction the outs and ins of test ints
c	never even went to server if this version of plinda is buffering
c       outs on the clients side. 
	call fxcommit(ftuple(1,frint(x,0)),12,"plinda.f")	

	call fxstart(12,"plinda.f")
	call fcreate_group(g,"testints", 12, "a.f")	
	call fout(g, ftuple(1,frint(int1,10)), 15, "plinda.f")
	call fout(g, ftuple(1,frint(int1,10)), 15, "plinda.f")
	call fout(g, ftuple(1,frint(int1,5)), 15, "plinda.f")
	call fxcommit(ftuple(1,frint(x,0)),12,"plinda.f")	

	call fxstart(12,"plinda.f")
	call fin(g, ftuple(1,flint(int2,10)), 16, "plinda.f")
	call fin(g, ftuple(1,flint(int2,10)), 16, "plinda.f")
	call fin(g, ftuple(1,flint(int2,10)), 16, "plinda.f")
	call fxcommit(ftuple(1,frint(x,0)),12,"plinda.f")	


	print *,"Testing transaction with stuff is done"

	return
	end


	subroutine testEval
	implicit none
	integer ftuple,frstr,fargrdp,frint,flchar
	character *20 myname
	integer ret,x

	ret = fargrdp(ftuple(1,flchar(myname,20)), 15, "plinda.f")
	if(ret .eq. 0)  then
c          this is the slave!! do not do infinite evaling
	   return
	end if

	myname = "test.exe"
	print *,"spawning another me"
	call fproc_eval(ftuple(2,frstr(myname),frint(x,0)),15,"plinda.f")

	return
	end


	subroutine testArgRdP
	implicit none
	integer ftuple,flchar,fargrdp,flint
	character *20 myname
	integer ret,x

	ret = fargrdp(ftuple(1,flchar(myname,20)), 15, "plinda.f")
	if(ret .eq. 0)  then
	   print *,"the current function is NOT!!! the master"
	else
	   print *,"this is the master its name = " , myname
	end if

	ret = fargrdp(ftuple(2,flchar(myname,20),flint(x,0)), 15, "plinda.f")
	if(ret .eq. 0)  then
	   print *,"the current function is NOT!!! the slave"
	else
	   print *,"this is the slave and its exec name = " , myname	
	end if
	return
	end


	subroutine testChars
	implicit none
	integer g(2),ftuple,flchar,frstr,frchar
	character *20 hello
	character *20,bye
	character h,b

	h = 'h'
	b = 'b'
	hello = "hello"
	bye = "bye"

	call fcreate_group(g,"testchars", 12, "a.f")

	call fout(g, ftuple(1,frchar(h,0)), 15, "plinda.f")
	call fin(g, ftuple(1,flchar(b,0)), 16, "plinda.f")

	if(h .ne. 'h' .or. b .ne. 'h') then
	   print *, "Error ***** testing strings 0"
	   return
	end if

	call fout(g, ftuple(1,frstr(hello)), 15, "plinda.f")
	call fin(g, ftuple(1,flchar(bye,20)), 16, "plinda.f")

	if(hello .ne. "hello") then
	   print *, "Error ***** testing strings 2"
	   return
	end if

	print *,"this should be the string hello <",bye,">"
	bye = "bye"


	call fout(g, ftuple(1,frchar(hello,20)), 15, "plinda.f")
	call fin(g, ftuple(1,flchar(bye,20)), 16, "plinda.f")

	if(hello .ne. "hello" .or. bye .ne. "hello") then
	   print *, "Error ***** testing strings 9"
	   return
	end if
	return
	end


	subroutine testInts
	implicit none
	integer int1(2),int2(2),g(2),ftuple,frint,flint
	integer int3,int4
        integer *8 int67,int68

	call fcreate_group(g,"testints", 12, "a.f")	

	int3 = 5
	int4 = 10
	call fout(g, ftuple(1,frint(int3,0)), 15, "plinda.f")
	call fin(g, ftuple(1,flint(int4,0)), 16, "plinda.f")
	if(int4 .ne. 5 .or. int3 .ne. 5) then
	   print *, "Error ***** testing scalar integers "
	   return
	end if
	
	int1(1) = 5
	int1(2) = 6
	int2(1) = 10
	int2(2) = 10

	call fout(g, ftuple(1,frint(int1,2)), 15, "plinda.f")
	call fin(g, ftuple(1,flint(int2,2)), 16, "plinda.f")

	if(int1(1) .ne. 5 .or. int1(2) .ne. 6) then
	   print *, "Error ***** testing int arrays 1"
	   return
	end if

	if(int2(1) .ne. 5 .or. int2(2) .ne. 6) then
	   print *, "Error ***** testing int arrays 2"
	   return
	end if

	int2(1) = 10
	int2(2) = 10

	call fout(g, ftuple(1,frint(int1,1)), 15, "plinda.f")
	call fin(g, ftuple(1,flint(int2,1)), 16, "plinda.f")

	if(int1(1) .ne. 5 .or. int1(2) .ne. 6) then
	   print *, "Error ***** testing int arrays 1"
	   return
	end if

	if(int2(1) .ne. 5 .or. int2(2) .ne. 10) then
	   print *, "Error ***** testing int arrays 6"
	   return
	end if


	return
	end



	subroutine testFloats
	implicit none
	integer g(2),ftuple,frfloat,flfloat
	real*4 int1(2),int2(2)
	real*4 int3,int4

	call fcreate_group(g,"testfloats", 12, "a.f")	

	int3 = 5.1
	int4 = 10.2

	call fout(g, ftuple(1,frfloat(int3,0)), 15, "plinda.f")
	call fin(g, ftuple(1,flfloat(int4,0)), 16, "plinda.f")

	if(int4 .ne. 5.1 .or. int3 .ne. 5.1) then
	   print *, "Error ***** testing scalar floats "
	   return
	end if
	
	int1(1) = 5.1
	int1(2) = 6.1
	int2(1) = 10.2
	int2(2) = 10.2

	call fout(g, ftuple(1,frfloat(int1,2)), 15, "plinda.f")
	call fin(g, ftuple(1,flfloat(int2,2)), 16, "plinda.f")

	if(int1(1) .ne. 5.1 .or. int1(2) .ne. 6.1) then
	   print *, "Error ***** testing float arrays 1"
	   return
	end if

	if(int2(1) .ne. 5.1 .or. int2(2) .ne. 6.1) then
	   print *, "Error ***** testing float arrays 4"
	   return
	end if

	int2(1) = 10.2
	int2(2) = 10.2

	call fout(g, ftuple(1,frfloat(int1,1)), 15, "plinda.f")
	call fin(g, ftuple(1,flfloat(int2,1)), 16, "plinda.f")

	if(int1(1) .ne. 5.1 .or. int1(2) .ne. 6.1) then
	   print *, "Error ***** testing float arrays 3"
	   return
	end if

	if(int2(1) .ne. 5.1 .or. int2(2) .ne. 10.2) then
	   print *, "Error ***** testing float 4"
	   return
	end if


	return
	end


	subroutine testDoubles
	implicit none
	integer g(2),ftuple,frdouble,fldouble
	real*8 int1(2),int2(2)
	real*8 int3,int4

	call fcreate_group(g,"testdoubles", 12, "a.f")	
	int3 = 5.1
	int4 = 10.2

	call fout(g, ftuple(1,frdouble(int3,0)), 15, "plinda.f")
	call fin(g, ftuple(1,fldouble(int4,0)), 16, "plinda.f")

	if(int4 .ne. 5.1 .or. int3 .ne. 5.1) then
	   print *, "Error ***** testing scalar doubles "
	   return
	end if
	
	int1(1) = 5.1
	int1(2) = 6.1
	int2(1) = 10.2
	int2(2) = 10.2

	call fout(g, ftuple(1,frdouble(int1,2)), 15, "plinda.f")
	call fin(g, ftuple(1,fldouble(int2,2)), 16, "plinda.f")

	if(int1(1) .ne. 5.1 .or. int1(2) .ne. 6.1) then
	   print *, "Error ***** testing double arrays 1"
	   return
	end if

	if(int2(1) .ne. 5.1 .or. int2(2) .ne. 6.1) then
	   print *, "Error ***** testing double arrays 4"
	   return
	end if

	int2(1) = 10.2
	int2(2) = 10.2

	call fout(g, ftuple(1,frdouble(int1,1)), 15, "plinda.f")
	call fin(g, ftuple(1,fldouble(int2,1)), 16, "plinda.f")

	if(int1(1) .ne. 5.1 .or. int1(2) .ne. 6.1) then
	   print *, "Error ***** testing double arrays 3"
	   return
	end if

	if(int2(1) .ne. 5.1 .or. int2(2) .ne. 10.2) then
	   print *, "Error ***** testing double arrays 41"
	   return
	end if


	return
	end


	subroutine testGids
	implicit none
	integer g(2),h(2),x(10),y(10),ftuple,frgid,flgid

	call fcreate_group(g,"testgids", 12, "a.f")	
	x(1) = g(1)
	x(2) = g(2)
	x(3) = 0
	x(4) = 0
	call fcreate_group(h,"testgids", 12, "a.f")	
	x(5) = h(1)
	x(6) = h(2)
	x(7) = 0
	x(8) = 0
	x(9) = 0
	x(10) = 0


	call fout(g, ftuple(1,frgid(g,0)), 15, "plinda.f")
	call fin(g, ftuple(1,flgid(h,0)), 16, "plinda.f")

	if(g(1) .ne. h(1) .or. g(2) .ne. h(2)) then
	   print *, "Error ***** testing scalar gids "
	   return
	end if
	
	call fout(g, ftuple(1,frgid(x,5)), 15, "plinda.f")
	call fin(g, ftuple(1,flgid(y,5)), 16, "plinda.f")

	if(x(1) .ne. y(1) .or. x(2) .ne. y(2)) then
	   print *, "Error ***** testing gid arrays 1"
	   return
	end if

	if(x(6) .ne. y(6) .or. x(4) .ne. y(4)) then
	   print *, "Error ***** testing gid arrays 1"
	   return
	end if

	if(x(1) .ne. g(1) .or. x(2) .ne. g(2)) then
	   print *, "Error ***** testing gid arrays 2"
	   return
	end if

	if(x(9) + x(10) + x(3) + x(4) .ne. 0) then
	   print *, "Error ***** testing gid arrays 3"
	   return
	end if


	return
	end




