
#include "W.h"
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <stdio.h>

typedef int boolean;
#define true 1
#define false 0

typedef struct {
  Display *display_;
  int screen_;
  Window window_;
  Drawable drawbuffer_;	/* For double buffering */
  Drawable stipple_;
  unsigned long foreground_;
  unsigned long background_;
  struct timeval *timer_;
  GC gc_;			/* Current gc */
  GC fg_;
  GC bg_;
  GC grayfg_;
  GC graybg_;
  XPoint points_[256];
  int point_count_;
  KeySym key_;
  int width_;
  int height_;
  boolean dirty_;
  boolean opaque_;
  long color_;
} WWindow;

static WWindow *WWindow__WWindow(int w, int h,char *winname);
static void WWindow__delete_WWindow(WWindow *);

static void WWindow__box(WWindow *, long ax, long ay, long bx, long by);
static void WWindow__vector(WWindow *, long ax, long ay, long bx, long by);
static void WWindow__color(WWindow *, long color);
static void WWindow__opacity(WWindow *, long opacity);
static void WWindow__sync(WWindow *);
static void WWindow__event(WWindow *, long *key, long *x, long *y, long *z);

static void WWindow__new_path(WWindow *);
static void WWindow__add_vertex(WWindow *, int x, int y);
static void WWindow__polygon(WWindow *);

static XEvent *WWindow__NextEvent(WWindow *, int *x, int *y,
				  unsigned int *state);
static Display *WWindow__Pending(WWindow *);
static void WWindow__Read(WWindow *, XEvent*, int *x, int *y,
			  unsigned int *state);
static void WWindow__GetChannels(WWindow *, fd_set *);
static struct timeval *WWindow__TimerInterval(WWindow *);

/* Zasha's own functions */
static void WWindow__SetPixel(WWindow *,int x,int y);
static void WWindow__Destroy (WWindow *);

static WWindow *default_window = 0;

/* INITIALIZATION */
void W_init(long width, long height,char *winname) {
  default_window = WWindow__WWindow(width, height,winname);}

/* GET THE NEXT INPUT EVENT */
void W_event(long *key, long *x, long *y, long *z) {
  WWindow__event(default_window, key, x, y, z);}

/* DRAW BOX ONTO INTERNAL IMAGE */
void W_box(long ax, long ay, long bx, long by) {
  WWindow__box(default_window, ax, ay, bx, by);}

/* DRAW VECTOR ONTO INTERNAL IMAGE */
void W_vector(long ax, long ay, long bx, long by) {
  WWindow__vector(default_window, ax, ay, bx, by);}

/* Zasha's */
void W_setpixel(int x,int y) {
  WWindow__SetPixel(default_window,x,y);}
void W_destroy (void) {
  WWindow__Destroy (default_window); }

void W_new_path(void) {
  WWindow__new_path(default_window);}
void W_add_vertex(int x, int y) {
  WWindow__add_vertex(default_window, x, y);}
void W_polygon(void) {
  WWindow__polygon(default_window);}

/* SET THE DRAWING COLOR */
void W_color(long color) {
  WWindow__color(default_window, color);}

/* SET THE DRAWING OPACITY */
void W_opacity(long opacity) {
  WWindow__opacity(default_window, opacity);}

/* UPDATE SCREEN FROM INTERNAL IMAGE */
void W_sync(void) {
  WWindow__sync(default_window);}

static WWindow *WWindow__WWindow(int w, int h,char *winname)
{
  XEvent *event;
  XSizeHints hint;
  WWindow *This = (WWindow*)malloc(sizeof(WWindow));
/*  WWindow *This = (WWindow*)new WWindow;
*/
  GC tmp_bg;			/* For drawing on stipple_, 1 plane deep */
  GC tmp_fg;			/* For drawing on stipple_, 1 plane deep */
  
  This->dirty_ = true;
  This->opaque_ = true;

  This->timer_ = (struct timeval *)malloc(sizeof(struct timeval));
/*  This->timer_ = new struct timeval; */
  This->timer_->tv_sec = 0;
  This->timer_->tv_usec = 1000030 / 60;

  This->display_ = XOpenDisplay("");
  This->screen_ = DefaultScreen(This->display_);
  This->background_ = WhitePixel(This->display_, This->screen_);
  This->foreground_ = BlackPixel(This->display_, This->screen_);
  This->color_ = This->foreground_;
  hint.width = w; hint.height = h;
  hint.flags = PSize;
  This->window_ = XCreateSimpleWindow(This->display_,
				      DefaultRootWindow(This->display_),
				      hint.x, hint.y, hint.width, hint.height,
				      5, This->foreground_, This->background_);
  XSetStandardProperties(This->display_, This->window_, winname, winname,
			 None, 0, 0, &hint);

  This->gc_ = This->fg_ = XCreateGC(This->display_, This->window_, 0, 0);
  XSetBackground(This->display_, This->fg_, This->background_);
  XSetForeground(This->display_, This->fg_, This->foreground_);

  This->bg_ = XCreateGC(This->display_, This->window_, 0, 0);
  XSetBackground(This->display_, This->bg_, This->foreground_);
  XSetForeground(This->display_, This->bg_, This->background_);

  This->stipple_ = XCreatePixmap(This->display_,This->window_, 2, 2, 1);
  tmp_fg = XCreateGC(This->display_, This->stipple_, 0, 0);
  XSetBackground(This->display_, tmp_fg, This->background_);
  XSetForeground(This->display_, tmp_fg, This->foreground_);
  tmp_bg = XCreateGC(This->display_, This->stipple_, 0, 0);
  XSetBackground(This->display_, tmp_bg, This->foreground_);
  XSetForeground(This->display_, tmp_bg, This->background_);
  
  XDrawPoint(This->display_, This->stipple_, tmp_fg, 0, 0);
  XDrawPoint(This->display_, This->stipple_, tmp_fg, 1, 1);
  XDrawPoint(This->display_, This->stipple_, tmp_bg, 0, 1);
  XDrawPoint(This->display_, This->stipple_, tmp_bg, 1, 0);
  This->grayfg_ = XCreateGC(This->display_, This->window_, 0, 0);
  This->graybg_ = XCreateGC(This->display_, This->window_, 0, 0);
  XSetBackground(This->display_, This->grayfg_, This->background_);
  XSetForeground(This->display_, This->grayfg_, This->foreground_);
  XSetBackground(This->display_, This->graybg_, This->foreground_);
  XSetForeground(This->display_, This->graybg_, This->background_);
  XSetFillStyle(This->display_, This->grayfg_, FillStippled);
  XSetStipple(This->display_,   This->grayfg_, This->stipple_);
  XSetFillStyle(This->display_, This->graybg_, FillStippled);
  XSetStipple(This->display_,   This->graybg_, This->stipple_);
  
  XSelectInput(This->display_, This->window_, 
	       PointerMotionMask |
	       ButtonPressMask | 
	       ButtonReleaseMask | 
	       KeyPressMask | 
	       ExposureMask);
  XMapRaised(This->display_, This->window_);
  
  This->point_count_ = 0;

  do {
    int x, y;
    unsigned int state;
    event = WWindow__NextEvent(This, &x, &y, &state);
  } while (!event || event->type != Expose);

  /* The first expose event gives us the real size, maybe?  Not really...*/
  This->width_ = event->xexpose.width;
  This->height_ = event->xexpose.height;
  This->drawbuffer_ = 
    XCreatePixmap(This->display_, This->window_, This->width_, This->height_,
		  DefaultDepth(This->display_, DefaultScreen(This->display_)));
  XFillRectangle(This->display_, This->drawbuffer_, This->bg_, 0, 0,
		 This->width_, This->height_);
  return This;

  /* Zasha */
  /* XSetFunction (This->display_,This->gc_,GX); */
  /* XSetForeground (This->display_,This->gc_,0xffffffff); */
}  

static void WWindow__delete_WWindow(WWindow *This)
{
  XFreeGC(This->display_, This->gc_);
  XDestroyWindow(This->display_, This->window_);
  /* XCloseDisplay(display_); */
}  


static int tmp;

#define Swap(a,b) (tmp=a,a=b,b=tmp)

static void WWindow__box(WWindow *This, long ax, long ay, long bx, long by)
{
  XSetFunction (This->display_,This->gc_,GXxor);
  if (ax > bx) Swap(ax,bx);
  if (ay > by) Swap(ay,by);
  XFillRectangle(This->display_, This->drawbuffer_, This->gc_, ax, ay, bx - ax, by - ay);
  This->dirty_ = true;
}


static void WWindow__vector(WWindow *This, long ax, long ay, long bx, long by)
{
  if (ax == bx && ay > by) Swap(ay,by);
  if (ay == by && ax > bx) Swap(ax,bx);
  XDrawLine(This->display_, This->drawbuffer_, This->gc_, ax, ay, bx, by);
  This->dirty_ = true;
}

static void WWindow__opacity(WWindow *This, long opacity)
{
	This->opaque_ = opacity;
	WWindow__color(This, This->color_);
}

static void WWindow__color(WWindow *This, long color)
{
  This->color_ = color;
  if (This->opaque_)
    if (color == This->background_) This->gc_ = This->bg_;
    else This->gc_ = This->fg_;
  else
    if (color == This->background_) This->gc_ = This->graybg_;
    else This->gc_ = This->grayfg_;
}

static void WWindow__sync(WWindow *This)
{
  if (This->dirty_) {
    XSetFunction(This->display_,This->gc_,GXcopy);
    XCopyArea(This->display_, This->drawbuffer_, This->window_, This->gc_,
	      0, 0, This->width_, This->height_, 0, 0);	    
    XFlush(This->display_);
    This->dirty_ = false;
  }
}

static void WWindow__new_path(WWindow *This)
{
  This->point_count_ = 0;
}

static void WWindow__add_vertex(WWindow *This, int x, int y)
{
  This->points_[This->point_count_].x = x;
  This->points_[This->point_count_].y = y;
  This->point_count_ += 1;
}

static void WWindow__polygon(WWindow *This)
{
  XFillPolygon(This->display_, This->drawbuffer_, This->gc_,
	       &This->points_[0], This->point_count_,
	       Complex, CoordModeOrigin);
  This->dirty_ = true;
}
 

static int time_click = 0;

static void WWindow__event(WWindow *This, long *key, long *x, long *y, long *z)
{
  unsigned char button;

  while (1) {
    int xi, yi;
    unsigned int state;
    XEvent *event = WWindow__NextEvent(This, &xi, &yi, &state);

    *x = xi; *y = yi; *z = 0;
    if (!event) {
      *key = TIME_CLICK;
      *z = time_click++;
      return;
    }
    else switch (event->type) {

    case MappingNotify:
      XRefreshKeyboardMapping((XMappingEvent*)&event);
      break;

    case ButtonPress:
      button = event->xbutton.button;
      if (button==3) {
	return;
      }
      *key = ((button == 1) ? LEFT_MOUSE_BUTTON :
	      (button == 2) ? MIDDLE_MOUSE_BUTTON :
	      (button == 3) ? RIGHT_MOUSE_BUTTON : -1);
      *z = 255;
      return;
      
    case ButtonRelease:
      button = event->xbutton.button;
      *key = ((button == 1) ? LEFT_MOUSE_BUTTON :
	      (button == 2) ? MIDDLE_MOUSE_BUTTON :
	      (button == 3) ? RIGHT_MOUSE_BUTTON : -1);
      *z = 0;
      return;

    case MotionNotify:
      state = event->xmotion.state;
      *key = MOUSE_MOTION;
      if (state & Button1Mask) {*z = 255; *key = LEFT_MOUSE_BUTTON;}
      else if (state & Button2Mask) {*z = 255; *key = MIDDLE_MOUSE_BUTTON;}
      else if (state & Button3Mask) {*z = 255; *key = RIGHT_MOUSE_BUTTON;}
      return;
      
    case KeyPress: {
      XKeyEvent *k = &event->xkey;
      char buf[4096];
      int len = XLookupString(k, buf, sizeof(buf), 0, 0);
      if (len != 0) {
	*key = buf[0];
	if (state & Mod1Mask) *key |= 0x80; /* Set the meta bit. */
	return;
      }
    }

    case Expose:
      printf("Expose event\n");
      This->dirty_=true;
      WWindow__sync(This);
#if 0
      if (event->xexpose.count == 0)
	XDrawImageString(event->xexpose.display,
			 event->xexpose.window,
			 w.gc_, 50, 50, hello, strlen(hello));
#endif
      break;
    }
  }
}

static XEvent *WWindow__NextEvent(WWindow *This, int *x, int *y,
				  unsigned int *state)
{
  static XEvent event;
  Display *display = WWindow__Pending(This);
  if (display) {
    WWindow__Read(This, &event, x, y, state);
    return &event;
  }
  else {
    fd_set channels;
    int r;
    FD_ZERO(&channels);
    WWindow__GetChannels(This, &channels);
    r = select(getdtablesize(), &channels, 0, 0, WWindow__TimerInterval(This));
    if (r == 0)
      return 0;
    else {
      WWindow__Read(This, &event, x, y, state);
      return &event;
    }
  }
}

static struct timeval *WWindow__TimerInterval(WWindow *This)
{
  return This->timer_ ? This->timer_ : 0;
}

static void WWindow__Read(WWindow *This, XEvent *event, int *x, int *y,
			  unsigned int *state)
{
  Window root, child;
  int root_x, root_y;
  XNextEvent(This->display_, event);
  if (event->type == MapNotify)
    while (XCheckTypedEvent(This->display_, MotionNotify, event));
  if (event->type == MotionNotify)
    while (XCheckTypedEvent(This->display_, MotionNotify, event));
  XQueryPointer(This->display_, This->window_, &root, &child,
		&root_x, &root_y, x, y, state);
}

static Display *WWindow__Pending(WWindow *This)
{
  XFlush(This->display_);
  if (QLength(This->display_) != 0)
    return This->display_;
  return 0;
}

static void WWindow__GetChannels(WWindow *This, fd_set *channels)
{
  FD_SET(XConnectionNumber(This->display_), channels);
}

/* Zasha's code follows */

static void WWindow__SetPixel (WWindow *This,int x,int y)
{
  XDrawPoint(This->display_,This->drawbuffer_,This->gc_,x,y);
  This->dirty_=true;
}

static void WWindow__Destroy (WWindow *This)
{
  XFreeGC(This->display_, This->gc_);

  XDestroyWindow(This->display_,This->window_);

  XCloseDisplay(This->display_);

  free(This);
}
