
// -------------------------------------------------------------
// matrix multiplication
//     Command line options:
//     -nN : NxN is the size of each matrix.
//     -sS : S = random number generator seed.                      
//     -c  : Check the result.
//     -o  : Print the matricies.
//     -h  : Print out command line options.
// -------------------------------------------------------------

#include <stdlib.h>
#include <string.h>
#include <iostream.h>
#include <sys/time.h>
#include <math.h>
#include "clock.C"

#ifdef  __GNUC__
#include <builtin.h>
extern "C" {
	double drand48();
	void srand48(long);
	int getopt(int , char * const *, const char *);
	extern char    *optarg;
}
#endif

const int	SIZE =	500;

int	size = SIZE;
long	seed = 0;
int	checkResult = 0;
int	outputResult = 0;

typedef	float	Matrix[SIZE][SIZE];

// -------------------------------------------------------------------------
void printUsage(char *prog_name) {
	cout << "Usage:\t" << prog_name << " <options>" << endl;
	cout << "\t-nN : NxN is the size of each matrix." << endl;
	cout << "\t-sS : S = random number generator seed." << endl;                    
	cout << "\t-c  : Check the result." << endl;
	cout << "\t-o  : Print the matricies." << endl;
	cout << "\t-h  : Print out command line options." << endl;
}

void getAndSetArgs(int argc, char *argv[]) {
	char		c;

	while ((c = getopt(argc, argv, "n:s:coh")) != -1) {
		switch(c) {
		   case 'n':	size = atoi(optarg);
				if (size > SIZE) {
					cout << "matricies must be smaller than ";
					cout << SIZE <<"X" << SIZE << "." << endl;
					exit(0);
				}
				break;
		   case 's':	seed = atoi(optarg);
				break;
		   case 'c':	checkResult = 1;
				break;
		   case 'o':	outputResult = 1;
				break;
		   case 'h':
		   default :	printUsage(argv[0]);
				exit(0);
		}
	}
}

// -------------------------------------------------------------------------
void randomFill(Matrix m) {
	for (int i=0; i<size; i++)
		for (int j=0; j<size; j++)
			m[i][j] = (drand48() - 0.5) * 1000.;
}

void print(Matrix m, int size) {
	cout.setf(ios::scientific, ios::floatfield);
	for (int i=0; i<size; i++) {
		for (int j=0; j<size; j++)
			cout << m[i][j] << "\t";
		cout << endl;
	}
}

const bool iszero(const float x)
{
   return x == 0.0;
}

int matrixEq(Matrix A, Matrix B, int rows, int cols) {
	for (int i=0; i<rows; i++)
		for (int j=0; j<cols; j++)
			if (!iszero(A[i][j] - B[i][j]))
				return 0;
	return 1;
}

void transpose(Matrix A, int size) {
	float		temp;
	for (int i=0; i<size; i++)
		for (int j=i+1; j<size; j++) {
			temp = A[i][j];
			A[i][j] = A[j][i];
			A[j][i] = temp;
		}
}

void mm(Matrix A, Matrix B, Matrix C, int dim1, int dim2, int dim3) {
	for (int i=0; i<dim1; i++) {
		for (int j=0; j<dim3; j++) {
			C[i][j] = 0.;
			for (int k=0; k<dim2; k++)
				C[i][j] += (A[i][k] * B[k][j]); 
		}
	}
}

void mm_cm(Matrix A, Matrix B, Matrix C, int dim1, int dim2, int dim3) 
{
   for (int i=0; i<dim1; i++) {
      for (int j=0; j<dim3; j++) {
	 C[i][j] = 0.;
	 for (int k=0; k<dim2; k++)
	    C[i][j] += (A[i][k] * B[j][k]); 
      }
   }
}
// -------------------------------------------------------------------------
	Matrix		A, B, C, D;

void main(int argc, char *argv[]) {
	getAndSetArgs(argc, argv);


	srand48(seed);
	randomFill(A);
	randomFill(B);

		if (checkResult)
			mm(A, B, D, size, size, size);

	cout << "Press return to start" << ::flush;
	char		c;
	cin.get(c);
	start("multiplication...");
	transpose(B, size);
	mm_cm(A, B, C, size, size, size);
	done();


	if (checkResult) {
		cout << "Checking the results..." << flush;
		if (matrixEq(C, D, size, size))
   		  cout << "succeeded." << endl << flush;

	}

	if (outputResult) {
		cout << "Matrix A:" << endl;
		print(A, size);
		cout << "Matrix B:" << endl;
		print(B, size);
		cout << "Matrix AXB:" << endl;
		print(C, size);
	}
}

