
#include "ds.h"
#include "Sl_fn.h"
#include "complex.h"

extern int num_locals;
extern int num_pots;

void
print_expn(s_complex* e)
{
    int i;
    for(i=0; i<PRECISION; i++){
	cout << "[" << e[i].rl << " " << e[i].im << "]";
    }
    cout << endl << flush;
}

/*------------------------------------------------------------------------------------*/
void
leaf_level_expns(int start_row, int grain_size, int *num_particles,
		 s_particle **particles,s_complex ***phi_val)
{

    int   r, i, prev_box, particles_in_box[LeafLevelDim+1];

    for(i=0; i<(dim[LeafLevel]+1); i++) 
	particles_in_box[i] = 0;
    for(r=0; r<grain_size; r++){
	/* for all rows */
	prev_box = 0;
	particles_in_box[prev_box] = 0;
	for(i=0; i<num_particles[r]; i++){
	    if (particles[r][i].n!=prev_box){
		/* i.e. if this particle doesnt belong to the prev box */
		prev_box++;
		particles_in_box[prev_box] = i;
	    }
	}
	prev_box++;
	
	particles_in_box[prev_box] = num_particles[r];
	for(i=0; i<dim[LeafLevel]; i++){
	    compute_multipole_expn(LeafLevel, start_row, r, i, particles_in_box[i], 
				   particles_in_box[i+1], particles, phi_val, 
				   Log2[grain_size]); 
	}
    }
    
}

/*------------------------------------------------------------------------------------*/
void
push_phi_values(int act_level,int act_row,int num_rows,int subtree_depth,
		s_complex ***phi_val)
{

    int  i, j, lev;
    s_complex  child_center, parent_center, by_dist;

    for(lev=subtree_depth; lev>0; lev--){
	for(i=0; i<num_rows; i++){
	    for(j=0; j<dim[act_level]; j++){
		child_center = get_center(act_level, act_row+i, j);
		parent_center = get_center(act_level-1, (act_row/2)+(i/2), j/2);
		by_dist = csub(child_center, parent_center);
		shift_phi(phi_val[lev][i] + j*PRECISION, 
			  phi_val[lev-1][i/2] + (j/2)*PRECISION, 
			  by_dist);
		/*shift  <child_expn> <parent_expn> <by_dist> */
	    }
	}
	num_rows /= 2;
	act_level--;
	act_row /= 2;
    }

}

/*------------------------------------------------------------------------------------*/
void
compute_ilist_values(int row,int level,int num_rows, 
		     s_complex ***phi_val,s_complex ***ilist_val)
{
    int i,j,k,p,d,real_i,real_p;

    for(i=0; i<(num_rows+4); i++)
      for(j=0; j<PRECISION*dim[level]; j++)
	   ilist_val[Log2[num_rows]][i][j] = NIL_COMPLEX;
    d = Log2[num_rows];
    for(i=0; i<num_rows; i += 2){
	for(j=0; j<dim[level]; j += 2){
	    for(k=j-2; k<j+4; k++){
		for(p=i-2; p<i+4; p++){
		    real_i = i + row;
		    real_p = p + row;
		    if (in_interaction(real_i,j,real_p,k,level,row))
			add_local_expansion(level,row,i+2,j,p+2,k,d,phi_val,ilist_val);
		    if (in_interaction(real_i+1,j,real_p,k,level,row))
			add_local_expansion(level,row,i+3,j,p+2,k,d,phi_val,ilist_val);
		    if (in_interaction(real_i,j+1,real_p,k,level,row))
			add_local_expansion(level,row,i+2,j+1,p+2,k,d,phi_val,ilist_val);
		    if (in_interaction(real_i+1,j+1,real_p,k,level,row))
			add_local_expansion(level,row,i+3,j+1,p+2,k,d,phi_val,ilist_val);
		}
	    }
	}
    }
}
/*------------------------------------------------------------------------------------*/
int 
in_interaction(int i,int j,int p,int k,int level,int row)
{

    if ((p<0) || (p>=dim[level]) || (k<0) || (k>=dim[level])) return 0;
    if (((p>=(i-1)) && (p<=(i+1))) && ((k<=(j+1)) && (k>=(j-1)))) return 0;
    return 1;
}
/*---------------------------------------------------------------------------------
  Compute the local expansion of the (p.k)_th box w.r.t. the center of the (i,j)_th
  box at level "l".
 -----------------------------------------------------------------------------------*/

void  
add_local_expansion(int level,int row,int i,int j,int p,int k,int depth,
		    s_complex ***phi_val,s_complex ***ilist_val)
{

   s_complex   tz, tmp_z, center_ij, center_pk, zp[PRECISION+1], exp_coeff[PRECISION+1];
   int  r,s, cK, cI;
   p_complex  boxj, boxk;

   //exp_coeff = new_expansion();
   for(r=0; r<PRECISION+1; r++) exp_coeff[r] = NIL_COMPLEX;
   tz = NIL_COMPLEX;

   num_locals++;
   /*-----------------------------------------------------
   if (level==LeafLevel){
       cout << "local = " << level << " row = " << row << " <" << i << " " << j << ">  <"
	   << p << " " << k << "> \n" << flush;
   }
   -----------------------------------------------------*/

   boxj = &(ilist_val[depth][i][j*PRECISION]);
   boxk = &(phi_val[depth][p][k*PRECISION]);

   center_ij = get_center(level, row+i, j);
   center_pk = get_center(level, row+p, k);

   tmp_z = csub(center_ij, center_pk);  /* dist between the two boxes */
   //zp = ;
   z_powers(zp,tmp_z);
   exp_coeff[0] = cmul(boxk[0],complex_log(cminus(tmp_z)));
   for(r=1; r<PRECISION; r++)
	if (r%2) exp_coeff[0] = csub(exp_coeff[0],czdiv(boxk[r],zp[r]));
	else     exp_coeff[0] = cadd(exp_coeff[0],czdiv(boxk[r],zp[r]));
   for(r=1; r<PRECISION; r++){
	tmp_z = czdiv(boxk[0],cimul(zp[r],r));
	for(s=1; s<PRECISION; s++)
	     if (s%2)  
		tz = csub(tz, cimul(czdiv(boxk[s],zp[s]),n_choos_k[r+s-1][s-1]));
	     else          
		tz = cadd(tz, cimul(czdiv(boxk[s],zp[s]),n_choos_k[r+s-1][s-1]));
	exp_coeff[r] = csub(czdiv(tz,zp[r]),tmp_z);
   }

   for(r=0; r<PRECISION; r++)
	boxj[r] = cadd(boxj[r], exp_coeff[r]);
   // delete [] zp;
}

/*--------------------------------------------------------------------------
  shift_phi : shifts the PHI value given w.r.t. the center of the (i,j)_th
  box on level "lev" to that w.r.t. the center of its parent box.
 ---------------------------------------------------------------------------*/

void  
shift_phi(s_complex* this_row,s_complex* parent_row,s_complex by_z)
{

   int          i, j;
   s_complex    Z, tmpz, zp[PRECISION+1];  

   //zp = 
   z_powers(zp,by_z);
   parent_row[0] = cadd(parent_row[0],this_row[0]);	
   for(i=1; i<PRECISION; i++){
        tmpz  = cmul(this_row[0],zp[i]);
	tmpz  = cdiv(tmpz,-i);
	for(j=1; j<=i; j++){
		Z = cmul(zp[i-j],this_row[j]);
		Z = cimul(Z,n_choos_k[i-1][j-1]);
	}
	parent_row[i] = cadd(Z,tmpz);
   }
   // delete [] zp;
}

/*--------------------------------------------------------------------------
  multipole_expansion takes the leaf level box # and computes the expansion 
  for PRECISION # of terms. Leaf level is globally defined.
  ---------------------------------------------------------------------------*/

void  
compute_multipole_expn(int level,int row,int offset,int box_no,
		       int  start_particle,int end_particle,
		       s_particle **particles,s_complex  ***phi_val, int depth)
{
    int          i,j,nump;
    s_complex    cZ,tZ,Z,Z1;
    p_particle   lp;
    
    cZ   = get_center(level, row+offset, box_no);	/* box center */
    lp   = particles[offset] + start_particle;
    nump = end_particle - start_particle;
    
    phi_val[depth][offset][box_no*PRECISION] = cons(total_charge(lp, nump),0.0);
    for (i=1; i<PRECISION; i++){
	Z = cons(0.0,0.0);
	for(j=0; j<nump; j++){
	    Z1 = csub(lp[j].z,cZ);
	    tZ = cfmul(cpower(Z1,i), lp[j].q);
	    Z  = cadd(tZ,Z);
        }
	phi_val[depth][offset][box_no*PRECISION + i] = cdiv(Z,-i);
    }
}
/*------------------------------------------------------------------------------
  Shift the local expansion of a distant box given w.r.t the parent box center
  to that w.r.t. the child box center.
 -------------------------------------------------------------------------------*/

void 
shift_local(s_complex by_z,s_complex *old_coeff,s_complex *new_coeff)
{
   s_complex zp[PRECISION+1];
   s_complex tz,z;
   int       l,k;

   //zp = 
   z_powers(zp, cminus(by_z));    /* precompute powers of shift in zp */
   for(l=0; l<PRECISION; l++){
	tz = cons(0.0,0.0);
	for(k=l; k<PRECISION; k++){
	  z  = cimul(*(old_coeff+k),n_choos_k[k][l]);
	  z  = cmul(z,zp[k-l]);
	  tz = cadd(tz,z);
        }
	*(new_coeff+l) = tz;
   }
   // delete [] zp;
}
/*--------------------------------------------------------------------------*/
float 
total_charge(s_particle *lp,int n)
{ 
    int i; float q;
    q = 0.0;
    for(i=0; i<n; i++) q += lp[i].q;
    return q;
}

/*--------------------------------------------------------------------------*/
void 
init_comb_table(){
    int i,j;
    for(i=0;i<2*PRECISION;i++){
	for(j=0; j<=i; j++){
	    n_choos_k[i][j] = fact[i]/(fact[j]*fact[i-j]);
	    if (j<i) n_choos_k[j][i] = 0;
	}
    }
}

/*--------------------------------------------------------------------------*/
s_complex  
get_center(int lev,int i,int j)
{

    float side;  s_complex c;
    
    side = RootBoxSide/dim[lev];
    c = cons(side*(i+0.5), side*(j+0.5));
    return c;
}
/*--------------------------------------------------------------------------*/
s_complex 
potential(int rowi,int pi,int rowj,int pj,s_particle **particles)
{
	
    s_complex f;
    f.rl = 0.0; f.im = 0.0;
    num_pots++;
    if (!csame(particles[rowi][pi].z,particles[rowj][pj].z)){
	f = cfmul(complex_log(csub(particles[rowi][pi].z,particles[rowj][pj].z)),
		  particles[rowj][pj].q);
    }
    else f = NIL_COMPLEX;
    return f;
}
/*-----------------------------------------------------------------------------
  Far field is computed by evaluating the "psi expansion" of each child box w.r.t.
  each particle position in the box.
 -------------------------------------------------------------------------------*/
s_complex 
eval_psi(s_complex* psi,s_complex z)
{
    s_complex f, zp[PRECISION+1]; 
    int i; 
    
    f = NIL_COMPLEX;
    //zp =
    z_powers(zp,z);
    for(i=0; i<PRECISION; i++)
	f = cadd(f, cmul(psi[i],zp[i]));
    // delete [] zp;
    return f;
}
/*--------------------------------------------------------------------------------*/
void 
compute_internal_fld(int offset,int  p_start,int  p_end,s_particle** particles)
{
    int j,k;
    
    if (p_end==(p_start+1)) return;
    for(j=p_start; j<p_end; j++){
	for(k=j+1; k<p_end; k++){
	    particles[offset][j].f = cadd(particles[offset][j].f, 
					  potential(offset,j,offset,k, 
						    particles));
	    particles[offset][k].f = cadd(particles[offset][k].f, 
					  potential(offset,k,offset,j, 
						    particles));
	}
    }
}
/*--------------------------------------------------------------------------------
 compute the psi values for the boxes in level lev_offset of the computation 
 subtree 
 -------------------------------------------------------------------------------*/
void
compute_psi(s_complex ***psi_val,s_complex ***ilist_val,s_complex ***psi_bar_val,
	    int level,int lev_offset)
{
    int i,j;

    /* for all rows on that level in the computation subtree */
    for(i=0; i<dim[lev_offset+1]; i++){
	/* for all boxes in each row ... */
	for(j=0; j<PRECISION*dim[level+lev_offset]; j++){
	    psi_val[lev_offset][i][j] = NIL_COMPLEX;
	}
    }

   for(i=0; i<dim[lev_offset+1]; i++){
	/* for all boxes in each row ... */
       for(j=0; j<PRECISION*dim[level+lev_offset]; j++){
	   psi_val[lev_offset][i][j] = cadd(psi_bar_val[lev_offset][i][j], 
					    ilist_val[lev_offset][i][j]);
       }
   }
}
/*----------------------------------------------------------------------
  compute psi_bar values for the next level given the values of psi for
  this level (basically we shift the psi values with respect to the centers
  of the child boxes of the boxes in this level)
 -----------------------------------------------------------------------*/
void
compute_psi_bar(s_complex ***psi_bar_val,s_complex ***psi_val,
		int start_row,int level,int lev_offset,int grain_size)
{

    /* we have to push the psi_values at level "lev_offset" as*/
    /* psi_bar_values at level "lev_offset + 1" */

    int  i, j, tmp1, tmp2;
    float  shift;
    s_complex   pc, by_z[4];
 
     /*cout << lev_offset << " -------------------------------------\n" << flush;
     if (lev_offset+1 ==Log2[grain_size]){
	 for(i=0; i<dim[lev_offset + 1]; i++){
	     for(j=0; j<dim[level + lev_offset]; j++){
		 cout << "(" << level+lev_offset << " " << 
		     i << " " << j << ") " ;
		 print_expn(&psi_val[lev_offset][i][j*PRECISION]);
	     }
	 }
     }*/

    for(i=0; i<grain_size && (i<dim[lev_offset+2]); i++)
      for(j=0; j<PRECISION*dim[level+lev_offset+1]; j++)
	   psi_bar_val[lev_offset+1][i][j] = NIL_COMPLEX;

    shift = RootBoxSide/(4.0*dim[level + lev_offset]);
    for(i=0; i<dim[lev_offset+1]; i++){
      for(j=0; j<dim[level + lev_offset]; j++){
	  pc       = get_center(level+lev_offset,start_row*dim[lev_offset+1] + i, j);
	  by_z[0]  = csub(pc, cons(pc.rl-shift, pc.im-shift));
	  by_z[1]  = csub(pc, cons(pc.rl+shift, pc.im-shift));
	  by_z[2]  = csub(pc, cons(pc.rl-shift, pc.im+shift));
	  by_z[3]  = csub(pc, cons(pc.rl+shift, pc.im+shift));
	  tmp1     = j*PRECISION;
	  tmp2     = 2*tmp1;
	  shift_local(by_z[0], psi_val[lev_offset][i] + tmp1, 
		      psi_bar_val[lev_offset + 1][2*i] + tmp2);
	  shift_local(by_z[1], psi_val[lev_offset][i] + tmp1, 
		      psi_bar_val[lev_offset + 1][2*i] + tmp2 + PRECISION);
	  shift_local(by_z[2], psi_val[lev_offset][i] + tmp1, 
		      psi_bar_val[lev_offset + 1][2*i+1] + tmp2);
	  shift_local(by_z[3], psi_val[lev_offset][i] + tmp1, 
		      psi_bar_val[lev_offset + 1][2*i+1] + tmp2 + PRECISION);

      }
  }
}
void 
compute_row_far_field(int row,int offset,s_particle** particles,int *num_particles,
		      s_complex***psi_val, int subtree_level)
{
   int i, bn;
   s_complex  box_center[LeafLevelDim], z, fld;

   for(i=0; i<LeafLevelDim; i++){
	box_center[i] = get_center(LeafLevel, row+offset, i);
    }
   for(i=0; i<num_particles[offset]; i++){
	bn  = particles[offset][i].n;
	z   = csub(particles[offset][i].z, box_center[bn]);
	fld = eval_psi(psi_val[subtree_level][offset] + bn*PRECISION, z);
	particles[offset][i].f = cadd(particles[offset][i].f, fld);
   }

}
/*--------------------------------------------------------------------------------
  for all boxes in the LeafLevel, which are offset from the row by "offset",
  compute the near field effect.
  --------------------------------------------------------------------------------*/
void 
compute_row_near_field(int row,int offset,s_particle **particles,
		       int *num_particles, int grain_size)
{
    int  i,j,k, p, prev_box, S, E,
	box[LeafLevelDim+1], box1[LeafLevelDim+1];
    
    
    prev_box = 0; 
    box[0]  = 0;
    box1[0] = 0;
    
    if (offset==0){
	for(i=0; i<num_particles[offset]; i++){
	    if (particles[offset][i].n!=prev_box){
		prev_box++;
		box[prev_box] = i;
	    }
	}
	prev_box++;
	box[prev_box] = num_particles[offset];
	for(i=0; i<(LeafLevelDim-1); i++) {
	    for(j=box[i]; j<box[i+1]; j++){
		for(k=box[i+1]; k<box[i+2]; k++){
		    particles[offset][j].f = 
			cadd(particles[offset][j].f, potential(offset,j,offset,k,particles));
		    particles[offset][k].f = 
			cadd(particles[offset][k].f, potential(offset,k,offset,j,particles));
		}
	    }
	    compute_internal_fld(offset, box[i], box[i+1],particles);
        }
	compute_internal_fld(offset, box[LeafLevelDim-1], box[LeafLevelDim],particles);
	/* compute the grain boundary values */
	/*if (row!=0){
	    box1[0] = 0;
	    prev_box = 0;
	    for(i=0; i<num_particles[grain_size]; i++){
		if (particles[grain_size][i].n!=prev_box){
		    prev_box++;
		    box1[prev_box] = i;
		}
	    }
	    prev_box++;
	    box1[prev_box] = num_particles[grain_size];
	    // interaction with previous column ... 
	    for(i=0; i<LeafLevelDim; i++){
		for(j=box[i]; j<box[i+1]; j++){
		    if (i==0) { S = 0; E = 1;  }
		    else if (i==(LeafLevelDim-1)) { S = i -1; E = i; }
		    else{ S = i-1; E = i+1;  }
		    for(p=S; p<=E; p++){
			for(k=box1[p]; k<box1[p+1]; k++){
			    particles[offset][j].f = 
				cadd(particles[offset][j].f, 
				     potential(offset,j,grain_size,k,particles));
			}
		    }
		}
	    }	    
	} */
    }
    else {
	for(i=0; i<num_particles[offset-1]; i++){
	    if (particles[offset-1][i].n!=prev_box){
		prev_box++;
		box[prev_box] = i;
	    }
	}
	prev_box++;
	box[prev_box] = num_particles[offset-1];
	prev_box = 0;
	for(i=0; i<num_particles[offset]; i++){
	    if (particles[offset][i].n!=prev_box){
		prev_box++;
		box1[prev_box] = i;
	    }
	}
	prev_box++;
	box1[prev_box] = num_particles[offset];

	/* same column computation .. */
	for(i=0; i<(LeafLevelDim-1); i++) {
	    for(j=box1[i]; j<box1[i+1]; j++){
		for(k=box1[i+1]; k<box1[i+2]; k++){
		    particles[offset][j].f = 
			cadd(particles[offset][j].f, potential(offset,j,offset,k,particles));
		    particles[offset][k].f = 
			cadd(particles[offset][k].f, potential(offset,k,offset,j,particles));
		}
	    }
	    compute_internal_fld(offset, box1[i], box1[i+1],particles);
	}
	compute_internal_fld(offset, box1[LeafLevelDim-1], box1[LeafLevelDim],particles);
	
	/* interaction with previous column ... */
	for(i=0; i<LeafLevelDim; i++){
	    for(j=box1[i]; j<box1[i+1]; j++){
		if (i==0) { S = 0; E = 1; }
		else if (i==(LeafLevelDim-1)) { S = i -1; E = i; }
		else{ S = i-1; E = i+1;	}
		for(p=S; p<=E; p++){
		    for(k=box[p]; k<box[p+1]; k++){
			particles[offset][j].f = 
			    cadd(particles[offset][j].f, potential(offset,j,offset-1,k,particles));
			particles[offset-1][k].f = 
			    cadd(particles[offset-1][k].f, potential(offset-1,k,offset,j,particles));
		    }
		}
	    }
	}
	/*
	  if ((row+grain_size)!=LeafLevelDim){
	    box[0] = 0;
	    // box1 computed above can be used here but not box[] 
	    prev_box = 0;
	    for(i=0; i<num_particles[grain_size+1]; i++){
		if (particles[grain_size+1][i].n!=prev_box){
		    prev_box++;
		    box[prev_box] = i;
		}
	    }
	    prev_box++;
	    box[prev_box] = num_particles[grain_size+1];
	    // interaction with next column ... 
	    for(i=0; i<LeafLevelDim; i++){
		for(j=box1[i]; j<box1[i+1]; j++){
		    if (i==0) { S = 0; E = 1;  }
		    else if (i==(LeafLevelDim-1)) { S = i -1; E = i; }
		    else{ S = i-1; E = i+1;  }
		    for(p=S; p<=E; p++){
			for(k=box[p]; k<box[p+1]; k++){
			    particles[offset][j].f = 
				cadd(particles[offset][j].f, 
				     potential(offset,j,grain_size+1,k,particles));
			}
		    }
		}
	    }	
         }    
         */
    }
}
/*--------------------------------------------------------------------------------*/
void
print_particles(s_particle** particles,int* num_particles,int  grain_size)
{
   int i,j,k;
   for(i=0; i<grain_size; i++){
      for(j=0; j<num_particles[i] && j<10; j++){
         cout << "<" << particles[i][j].n << " " << particles[i][j].f.rl << ">";
      }
      cout << endl << flush;
   }
}
/*--------------------------------------------------------------------------------*/








