
#include <stdio.h>
#include <malloc.h>
#include "ds.h"
#include "complex.h"
#include "plinda_fns.h"
#include "Sl_fn.h"

//extern "C" {
//void * calloc(long unsigned int, long unsigned int);
//}


/* ------------------------------- Definitions ------------------------------------ */

get_job(int *pass,int *start_row,int *level,int grain_size, gid g)
{
    int cur_pass, cur_level, num_done, num_todo, lev, row, i, next_level;

    in_counter(cur_pass, cur_level, num_done, num_todo, g);
    if (cur_pass==JobsFinished){
	out_counter(cur_pass,cur_level,num_done,num_todo, g);
	return -1;
    }
    if (cur_pass==UpwardPass){
	if (cur_level<=1){
	    /* upward pass done . start the next pass*/
	    /* all jobs for middle pass are generated now */
	    lev = 1; row =0; num_todo = 0;
	    while(1){
		if (dim[lev] <= row){
		    lev++; row = 0; 
		    if (lev>LeafLevel) break;
		    continue;
		}
		out_job(lev, row, g);
		num_todo++;
		row += grain_size;
	    }
	    /* job counter for the middle pass */
	    out_counter( MiddlePass, 1, 0, num_todo, g);
	    return 0;
	}
	else if (num_done==num_todo){
	    next_level = cur_level - Log2[grain_size];
	    if (grain_size==1) 
		next_level = cur_level - 1;
	    if (next_level<=1){
		cout << "upward pass finished. \n" << flush;
		/* upward pass done . start the next pass*/
		/* all jobs for middle pass are generated now */
		lev = 1; row =0; num_todo = 0;
		while(1){
		    if (dim[lev] <= row){
			lev++; row = 0; 
			if (lev>LeafLevel) break;
			continue;
		    }
		    out_job(lev, row, g);
		    num_todo++;
		    row += grain_size;
		}
		/* job counter for the middle pass */
		out_counter( MiddlePass, 1, 0, num_todo, g);
		return 0;
	    }
	    else {
		num_todo = 0;
		/*create jobs for that level */
		for(i=0; i<dim[next_level]; i += grain_size){
		    out_job(next_level, i, g);
		    num_todo++;
		}
		out_counter( cur_pass, next_level, 0, num_todo, g);
		return 0;
	    }
	}
	else{
	    /* there are jobs to be done this level */
	    in_job01(cur_level,row, g);
	    *pass = cur_pass;
	    *start_row = row;
	    *level = cur_level;
	    num_done++;
	    out_counter( cur_pass, cur_level, num_done, num_todo, g);
	    return 1;
	}
    }
    if (cur_pass==MiddlePass){
	if (num_todo==num_done){
	    cout << "middle pass done...\n" << flush;
	    /* middle pass finished . init for downward pass */
	    out_counter( DownwardPass, 1, 0, 1, g);
	    out_job(1, 0, g);
	    return 0;
	}
	in_job11(cur_level, row, g);
	*pass = cur_pass;
	*start_row = row;
	*level = cur_level;
	num_done++;
	out_counter( cur_pass, cur_level, num_done, num_todo, g);
	return 1;
    }
    else if (cur_pass==DownwardPass){
	if ((cur_level>LeafLevel) || (cur_level==LeafLevel && num_done==num_todo)){
	    /* all tasks finished */
	    out_counter( JobsFinished, next_level, 0,0, g);
	    return -1;
	}
	else if (num_done==num_todo){
	    /* all jobs in this level done */
	    /* start computation in the next level */
	    next_level = cur_level + Log2[grain_size];
	    if (next_level>=LeafLevel) {
	        out_counter( JobsFinished, next_level, 0,0, g);
                return -1;
            }
	    out_counter( cur_pass, next_level,0, dim[next_level], g);
	    for(i=0; i<dim[next_level]; i++){
		out_job(next_level, i, g);
	    }
	    return 0;
	}
	else{
	    /* there are jobs in this level */
	    in_job01(cur_level, row, g);
	    *pass = cur_pass;
	    *start_row = row;
	    *level = cur_level;
	    num_done++;
	    out_counter( cur_pass, cur_level, num_done, num_todo, g);
	    return 1;
	}	    
    }
    else{
	/* error case */
	out_counter( cur_pass, cur_level, num_done, num_todo, g);
	return -1;
    }

}
/*-------------------------------------------------------------------------*/
int
get_data(int pass,int start_row,int level,int grain_size,
	 s_particle **particles,int *num_particles, 
	 s_complex ***phi_val,s_complex ***psi_val, 
	 s_complex ***psi_bar_val,s_complex ***ilist_val, gid g)
{

    int  i, j, k, row;

    if (pass==UpwardPass){
	if (level==LeafLevel){
	    /* read particle data */
	    for(i=0; i<grain_size && ((i+start_row) < dim[level]); i++)
		rd_particles(i+start_row, num_particles[i], particles[i], g);
	}
	else{
	    /* read the phi values */
	    /* subtree_lev = ((start_row+grain_size)>=dim[level]) ? 
	       dim[level] - start_row : grain_size; 
	       It is sort of complicated - so we restrict to case when grain_size 
	       chosen such that every subtree is proper and identical to 
	       evryother .. */
	    for(i=0; i<grain_size && (start_row+i < dim[level]); i++){
		/* read the values  into the bottom most level */
		in_phi(level, i+start_row, phi_val[Log2[grain_size]][i], g);
		/*for(k=0; k<dim[level]; k++){
		    cout << "<" << level << " " << i << " " << k << "> ";
		    print_expn(&phi_val[Log2[grain_size]][i][k*PRECISION]);
		}*/
	    }
         }
	return 1;
    }
    else if (pass==MiddlePass){
	for(j=0; j<dim[LeafLevel]*PRECISION; j++){
	    phi_val[Log2[grain_size]][0][j] = NIL_COMPLEX;
	    phi_val[Log2[grain_size]][1][j] = NIL_COMPLEX;
	    phi_val[Log2[grain_size]][grain_size][j] = NIL_COMPLEX;
	    phi_val[Log2[grain_size]][grain_size+1][j] = NIL_COMPLEX;
	}
	/* read the phi values */
	for(i=-2; i <= (grain_size+1); i++){
	    /* the two boundary rows are needed to compute the ilist values
	       for the edge rows.. */
	    if (((start_row+i) <0) || ((start_row+i)>=dim[level]))
		continue;
	    rd_complex_array("phi", level, i+start_row, phi_val[Log2[grain_size]][i+2],
			     g);
	}
	return 1;
    }
    else if (pass==DownwardPass){
	/* got to read the whole subtree of ilist_values and 
	   the value of psi_bar of the root row of the subtree */
	row = start_row;
	for(i=0; i<=Log2[grain_size]; i++){
	    /* for each level .. */
	    for(j=0; j<dim[i+1]; j++){
		/* read the ilist values at that level */
		rd_complex_array("ilist", level + i, j+row, ilist_val[i][j],
				 g);
	    }
	    row *= 2;
	}
	/* read the root psi_bar values . the psi values can be computed
	 from the psi and the ilist values.*/
	if (level!=1){
	    rd_complex_array("psi_bar", level,start_row,psi_bar_val[0][0], g);
	    /*for(k=0; k<dim[level]; k++){
		cout << "==[" << level << " " << start_row << " " << k<< "]== ";
		print_expn(&psi_bar_val[0][0][k*PRECISION]);
	    }*/
	}	    
        else{
	    for(i=0; i<dim[LeafLevel]*PRECISION; i++)
	        psi_bar_val[0][0][i] = NIL_COMPLEX;
        }
	/* in the particle values if the current subtree is the lower
	   most */
	row = start_row*grain_size;
	if (level+Log2[grain_size] == LeafLevel){
	    for(i=1; i<grain_size-1; i++){
		in_particles(i+row, num_particles[i], particles[i], g);
	    }
	    i = 0;
	    in_particles(i+row, num_particles[i], particles[i], g);
	    i = grain_size-1;
	    in_particles(i+row, num_particles[i], particles[i], g);
	    /*if (row!=0){
		rd_particles(row-1, num_particles[grain_size], 
			     particles[grain_size], g);
	    }
	    if ((row+grain_size)!= LeafLevelDim){
		rd_particles(row+grain_size, num_particles[grain_size+1], 
			     particles[grain_size+1], g);
	    }*/
	}
	return 1;
    }
    return 0;

    /* note that since we never readin psi values, they never need to be 
       dumped into the tuple space. it is only the psi bar values of the 
       leaf boxes of the subtrees and the ilist values of all the boxes 
       that matter. */
}

/*--------------------------------------------------------------------------------*/
int
compute(int pass,int start_row,int level,int grain_size,
	 s_particle **particles,int *num_particles, 
	 s_complex ***phi_val,s_complex ***psi_val, 
	 s_complex ***psi_bar_val,s_complex ***ilist_val)
{
    int i,j,k, row;

    if (pass==UpwardPass){
	/* data we have : if leaf level : the particle data at the bottom 
	   else phi values of the base of the subtree whose phi values are
	   to be computed */
	if (level==LeafLevel){
	    leaf_level_expns(start_row, grain_size, num_particles, 
			     particles, phi_val);
	}
	push_phi_values(level, start_row, grain_size,Log2[grain_size], phi_val);
    }
    else if (pass==MiddlePass){
	/* no ilist values at the top two levels */
	if (level==1 || level==2){
            for(j=0; j<grain_size+2; j++){
	       for(i=0; i<dim[LeafLevel]*PRECISION; i++){
	           ilist_val[Log2[grain_size]][j][i] = NIL_COMPLEX;
               }
            }
            return 1;
        }
	compute_ilist_values(start_row, level, grain_size, phi_val, ilist_val);
    }
    else{ 
	/* DownwardPass */
	for(i=0; i<=Log2[grain_size]; i++){
	    if (level+i > LeafLevel) break;
	    compute_psi(psi_val, ilist_val, psi_bar_val, level, i);
	    if (i==Log2[grain_size]) break;
	    compute_psi_bar(psi_bar_val, psi_val, start_row, level,i, grain_size);

	    /*for(j=0; j<dim[i+1]; j++){
		for(k=0; k<dim[level]; k++){
		    cout << "<" << level+i << " " << j + start_row*dim[i+1] << " " << k << "> ";
		    print_expn(&psi_bar_val[i][j][k*PRECISION]);
		}
	    }*/

	}
	if ((level+Log2[grain_size])>=LeafLevel){
	    row = start_row*dim[LeafLevel - level + 1];
	    j = ((LeafLevel-level)<Log2[grain_size]) ? (LeafLevel - level) : Log2[grain_size];
	    /* need to compute the far field and near field effects */
	    for(i=0; i<grain_size && ((row + i)<dim[LeafLevel]); i++){
		compute_row_far_field(row, i,particles,num_particles,psi_val,j);
		/*print_particles(particles,num_particles,grain_size);*/
		compute_row_near_field(row, i,particles,num_particles,grain_size);
	    }
	    /*cout << "\n****************************************\n" << flush;
	    print_particles(particles,num_particles,grain_size);
	    cout << "\n=======================================\n" << flush;*/
	}
    }

}

/*--------------------------------------------------------------------------------*/

output_results(int pass,int start_row,int level,int grain_size,
	       s_particle **particles,int *num_particles, 
	       s_complex ***phi_val,s_complex ***psi_val, 
	       s_complex ***psi_bar_val,s_complex ***ilist_val, gid g)
{

    int  row, lev, i, j, k;

    if (pass==UpwardPass){
	/* output the whole subtree of phi values */
	row = start_row;
	lev = level;
	for(i=Log2[grain_size]; i>=0; i--){
	    /* for each level */
	    for(j=0; j<dim[i+1]; j++){
		/* all rows in that level */
		out_complex_array("phi", lev, row + j, phi_val[i][j], g);
		/*for(k=0; k<dim[lev]; k++){
		    cout << "<" << lev << " " << j << " " << k << "> ";
		    print_expn(&phi_val[i][j][k*PRECISION]);
		}*/
		for(k=0; k<dim[lev]*PRECISION; k++){
		    phi_val[i][j][k] = NIL_COMPLEX;
		}
	    }
	    row /= 2;
	    lev--;
	}
    }
    else if (pass==MiddlePass){
	/* output all the ilist values of the "relevant" rows */
	for(i=0; i<grain_size && (i+start_row)<dim[level]; i++){
	    out_complex_array("ilist", level, i+start_row, 
			      ilist_val[Log2[grain_size]][i+2], g);
	   /* if (level==LeafLevel) {
		for(k=0; k<dim[level]; k++){
		    cout << "<" << level << " " << i+start_row << " " << k << "> ";
		    print_expn(&ilist_val[Log2[grain_size]][i+2][k*PRECISION]);
		}
	    }*/
	}
    }
    else{ /* DownwardPass */
	/*  output the psibar values of the leaf level rows of
	    the computation subtree. Incase of the leaf level rows
	    are also the leaf level rows of the actual computation 
	    tree, then output the particle data too */
	row = start_row*grain_size;
	lev = level + Log2[grain_size];
	for(i=0; i<grain_size && (i+row)<dim[lev]; i++){
	    out_complex_array("psi_bar", lev, i+row, psi_bar_val[Log2[grain_size]][i],
			      g);
	    /*if (lev==LeafLevel && ((i+row)==(dim[lev]-1))){
		for(k=0; k<dim[lev]; k++){
		    cout << "[" << lev << " " << i+row << " " << k << "] ";
		    print_expn(&psi_bar_val[Log2[grain_size]][i][k*PRECISION]);
		}
	    }*/
	}
	if ((level + Log2[grain_size])==LeafLevel){
	    /* the particle data too ... 
	       print_particles(particles,num_particles,grain_size);*/
	    /*in_particles(row, num_particles[grain_size], 
			 particles[grain_size], g);
	    in_particles(row+grain_size-1, num_particles[grain_size+1], 
			 particles[grain_size+1], g);
	    */
	    for(i=0; i<grain_size && (i+row)<dim[lev]; i++){
		out_particles(i+row, num_particles[i], particles[i], g);
	    }
	}
    }
}
/*--------------------------------------------------------------------------------*/
void
alloc_memory(int grain_size, s_complex ****phi_val, s_complex ****psi_val, 
	     s_complex ****psi_bar_val, s_complex ****ilist_val,
	     int **num_particles,s_particle ***particles)
{
    int i,j,tmp;

    (*num_particles) = (int*)calloc(grain_size+2,sizeof(int));
    (*particles)     = (s_particle**)calloc(grain_size+2,4);
    for(i=0; i<grain_size+2; i++)
	(*particles)[i] = (s_particle*)calloc(MaxNumRowPar, sizeof(s_particle));

    (*phi_val)     = (s_complex***)calloc(Log2[grain_size]+1,4); 
    (*psi_val)     = (s_complex***)calloc(Log2[grain_size]+1,4);
    (*psi_bar_val) = (s_complex***)calloc(Log2[grain_size]+1,4);
    (*ilist_val)   = (s_complex***)calloc(Log2[grain_size]+1,4);

    for(i=0; i<=Log2[grain_size]; i++){
	(*phi_val)[i] = (s_complex**)calloc(dim[i+1]+4, 4);
	(*psi_val)[i] = (s_complex**)calloc(dim[i+1]+4, 4);
	(*psi_bar_val)[i] = (s_complex**)calloc(dim[i+1]+4, 4);
	(*ilist_val)[i] = (s_complex**)calloc(dim[i+1]+4, 4);
    }
    tmp = sizeof(s_complex);
    for(i=0; i<=Log2[grain_size]; i++){
	for(j=0; j<(dim[i+1]+4); j++){
	    (*phi_val)[i][j] = (s_complex*)calloc(dim[LeafLevel]*PRECISION, tmp);
	    (*psi_val)[i][j] = (s_complex*)calloc(dim[LeafLevel]*PRECISION, tmp);
	    (*psi_bar_val)[i][j] = (s_complex*)calloc(dim[LeafLevel]*PRECISION, tmp);
	    (*ilist_val)[i][j] = (s_complex*)calloc(dim[LeafLevel]*PRECISION, tmp);
	}
    }

}
/*--------------------------------------------------------------------------------*/
