/*------------------------------------------------------------------------------
         			<< complex.c >>

   Functions to manipulate complex numbers...
  ------------------------------------------------------------------------------*/

#include <math.h>
#include <stdio.h>
#include "ds.h"


int czero(s_complex z) 
{ return ( (z.rl==0.0) && (z.im==0.0)); }

s_complex cons(float r,float i) 
{ s_complex z; z.rl = r; z.im = i; return z; }

s_complex cmul(s_complex z1,s_complex z2) 
{ s_complex tz;  
     tz.rl = z1.rl*z2.rl - z1.im*z2.im; 
     tz.im = z1.rl*z2.im + z1.im*z2.rl; return tz; }

s_complex cpower(s_complex z,int i) 
{ int j; s_complex tz;  tz.rl = 1.0; tz.im = 0.0;
     if (i==0) return tz;
     tz = z; for(j=1; j<i; j++) tz = cmul(tz,z); return tz; }

s_complex csub(s_complex z1,s_complex z2) 
{ s_complex tz; 
     tz.rl = z1.rl - z2.rl; tz.im = z1.im -  z2.im; return tz; }

s_complex cadd(s_complex z1,s_complex z2) 
{ s_complex tz; 
     tz.rl = z1.rl + z2.rl; tz.im = z1.im +  z2.im; return tz; }

s_complex cdiv(s_complex z,int i) 
{ s_complex tz; 
     tz.rl = z.rl / ((float)i); tz.im = z.im / ((float)i); return tz; }

s_complex cfmul(s_complex z,float i) 
{ s_complex tz; 
     tz.rl = z.rl * i; tz.im = z.im *i; return tz; }

s_complex cimul(s_complex z,int i) 
{ s_complex tz; 
     tz.rl = z.rl*i; tz.im = z.im*i; return tz; }

s_complex czdiv(s_complex z1,s_complex z2) 
{ s_complex tz; float f = 0.0;
     if (czero(z2)) { cout << "Div by zero complex\n" << flush; }
     f = z2.rl*z2.rl + z2.im*z2.im;
     tz.rl = (z1.rl*z2.rl + z1.im*z2.im)/f;
     tz.im = (z1.im*z2.rl - z1.rl*z2.im)/f;
     return tz; }

s_complex cminus(s_complex z) 
{ s_complex tz; 
     tz.rl = - z.rl; tz.im = - z.im; return tz; }

s_complex complex_log(s_complex z) 
{ s_complex tz; float x,y,theta,f;
     tz.rl = 0.0; tz.im = 0.0;
     x = z.rl; y = z.im; f = (float)sqrt(x*x + y*y);
     if (x==0.0){
	if (y>0.0) theta = 3.141*0.5; else theta = -3.141*0.5; }
     else theta = atan (y/x);
     if (f==0.0) { (int) printf("Log of zero ?? \n");}
     tz.rl = (float)log((double)f); tz.im = theta;
     return tz;
   }

int csame(s_complex z1,s_complex z2) 
{  return ((z1.rl==z2.rl) && (z1.im==z2.im)); }

p_complex new_expansion()
{  p_complex t;
   t = (p_complex)calloc(PRECISION+1,sizeof(s_complex));
   return t;
}


/*p_complex */
void
z_powers(s_complex *zp, s_complex z)
{  int i; /*p_complex zp;*/
   /*zp = new_expansion();*/
   for(i=0; i<PRECISION; i++) *(zp+i) = cpower(z,i);
   /*return zp;*/
   return;
}

/*-----------------------------------------------------------------------*/
