// This may look like C code, but it is really -*- C++ -*-
//
// File:        Predictor.C
// Description:	functions for predicting expected times
// Created:  
// Author:      Hansoo Kim/Peter Wyckoff
// Mail:        hansoo@cs.nyu.edu/wyckoff@cs.nyu.edu
//
#include <stream.h>
#include "Predictor.h"

const double cc_base = 0.0003248;
const double cc_k = 1.13904e-07;
const double in_base = 0.0003916;
const double in_k = 1.50253e-07;
const double rd_base = 0.0003894;
const double rd_k = 1.1562e-07;
const double out_base = 2.4e-05;
const double out_k = 1.24849e-07;
const double xr_base = 0.0003894;
const double xr_k = 1.1562e-07;
const double log_base = 1.6e-06;
const double log_k = 2.16933e-08;
const double bcopy_base = 8.6e-06;
const double bcopy_k = 1.45734e-08;
const double mcopy_base = 4.8e-06;
const double mcopy_k = 1.13163e-08;
const double proc_recovery_time =1.00244;

double expectedTime(double failProb, double failFreeTime, double recoverTime)
{
  // based on the recurrence
  // T(N) = (1-p) (T(N-1)+1)  + p (T(N) + R)
  // solution is: N * (1-p+pr)/(1-p)
  return failFreeTime * (1 - failProb + failProb * recoverTime)/(1-failProb);
}

double 
commitConsistentTime(double secs,int commit_size,int repeat,double alpha)

{
  const double tranTime = secs + cc_base + cc_k * commit_size;
  const double basicRecovery = 
    proc_recovery_time + xr_base + xr_k * commit_size + tranTime/2;
  return expectedTime(alpha, repeat * tranTime, basicRecovery);

}

double messageReplayTime(double compute_time,int commit_size, 
			 int repeat, unsigned  interval, double alpha,
			 int rdsNum,int rdsLength,
			 int outsNum,int outsLength)
{
  const double logMaintenance = rdsNum * mcopy_base + rdsLength * mcopy_k;
  const double tranTime = compute_time + cc_base + logMaintenance;
  const double numTransPerInterval = interval/tranTime;
  const double basicRecovery = 
    proc_recovery_time + xr_base + xr_k * commit_size + tranTime/2;
  const double sendingLog = 
    xr_k * rdsLength * numTransPerInterval/2;
  const double tranTimeReplay = compute_time - rdsNum * rd_base -
    rdsLength * rd_k - outsNum * out_base - outsLength * out_k;
  const double stateReconstruction = 
    numTransPerInterval/2 * tranTimeReplay;
  const double r_ml = basicRecovery + sendingLog + stateReconstruction;
  return expectedTime(alpha, repeat * tranTime, r_ml);
}





double triggeredTime(double compute_time,int commit_size, 
		     int repeat, unsigned  interval, double alpha,
		     int numProcs, double totContinuations )
{
  const double tranTime = compute_time + cc_base;
  const double globalReconstruct = 
    numProcs * (proc_recovery_time + xr_base + totContinuations * xr_k +
		interval/2);

  return expectedTime(alpha, repeat * tranTime, globalReconstruct);
}




