/*********
File:			Analyze_for_Emittance.c
Purpose:		This file analyzes the PLinda code and calls emittance
			functions in file Emit_PLinda.c to output it into a file
			for later compilation/linking.
Last Altered:		11/24/96
Author:			Tom Brown(brow7141@cs.nyu.edu)
Project:		PLinda(Version 2.6) Fortran 77
*********/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "SymTable.h"
#include "Match_keywords.h"
#include "plinda_f.h"
#include "Match_Tools.h"
#include "Char_Match.h"
#include "List.h"
#include "Analyze_for_Emittance.h"
#include "Emit_PLinda.h"

#define FLOAT 1
#define ASCII_ZERO 48
#define NO_FURTHER_EMITTANCE_NEEDED 5

int count_args;




/******
Function:	Check_for_GID
Purpose:	Returns 1 if a 'gid' argument is specified for a PLinda
		function call; returns 0, otherwise.
Called From: 	Immediate_Stateless_Emission()
******/
char *
Check_for_GID(SymTable * Sym_Table, CONTEXT * c)
{
	List * var_list; 
	int i = 0;
	char * gid_name, * front, hold;
	gid_name = (char *)malloc(sizeof(char) * 80);
	front = gid_name;

/**
	fprintf(stderr,"Check_for_GID::  ");
	CON_print(c);
**/
	do {
		*gid_name = CON_getc(c);
		if (End_of_Line(*gid_name)) {
			return (NULL);
		}
	} while (*gid_name != '[');


	hold = *gid_name;
	while (hold != ']') {
		*gid_name = CON_getc(c);
		hold = *gid_name;
		gid_name++;
		if (End_of_Line(hold)) {
			fprintf(stderr,"Error:: Line %d:  ",c->line_number);
			CON_print(c);
			return (NULL);
		}
	}
	gid_name--;
	gid_name[i] = '\0';

/**
	fprintf(stderr,"Check_for_GID:: gid_name => %s\n",front);
**/

	var_list = Access_List(Sym_Table->list, front);
	if (var_list == NULL) {
		Print_SymTable(Sym_Table);
		fprintf(stderr,"Error:: Line %d: ",c->line_number);
		fprintf(stderr,"'%s' undeclared gid.\n", front);
		CON_print(c);
		fprintf(stderr,"\n");
		return (NULL);
	} else {
	/**	fprintf(stderr,"var_type => %d\n", var_list->variable_type);
		fprintf(stderr,"Error:: Line %d: ",c->line_number);
		fprintf(stderr,"'%s' undeclared gid.\n", front);
	**/
		if (!((var_list->variable_type == INTEGER) ||
				(var_list->variable_type == GID))) {
			fprintf(stderr,"Error:: Line %d: '%s' not of type gid.\n", c->line_number, front);
			CON_print(c);
			return (NULL);
		}
	}
	return (front);
}



/******
Function:	Is_Tuple_Space_Empty
Purpose:	Returns 1 if the there are no tuple space arguments.
		Returns 0, otherwise.
******/
int
Is_Tuple_Space_Empty(CONTEXT * c)
{
	char temp;
	CON_Align_to_Char('(', c);
	do {
		temp = CON_getc(c);
		if (End_of_Line(temp)) {
			return 1;
		}
/**
		fprintf(stderr,"Is_Tuple_Space_Empty:: char => %c \n",temp);
**/
		if ((White_Space(temp) == 0) && (temp != ')'))
			return 0;
	} while (temp != ')');
/**
	fprintf(stderr,"\n\nIs_Tuple_Space_Empty:: No!! Return 1 !!! \n\n");
**/
	return 1;
}



/******
Function:	Immediate_Stateless_Emission
Purpose:	Emit PLinda functions according to parsing_mode and/or parameter
		arguments for tuple-space.  Returns 1, if the function emission
		is 'stateless' and needs no further tuple-space interpretation.
		Otherwise, returns 0.
******/
int
Immediate_Stateless_Emission(int parsing_mode, SymTable * Sym_Table, CONTEXT * c, FILE * plinda)
{
   char * gid_name;

/**	1) Check if gid specified.  If not, then the PLinda function 
		call has no tuple space arguments, thus output function
		name and exit. Otherwise, output PLinda function name. **/
   gid_name = Check_for_GID(Sym_Table, c);
   if (gid_name == NULL) {
	if (parsing_mode == XCOMMIT) {
		if (!Is_Tuple_Space_Empty(c)) {
		   Emit_Sundry_PLinda_calls(parsing_mode, c, plinda);
		   return 0;
		}
	}
	if ((parsing_mode == XSTART) 	||
	    (parsing_mode == DESTROY_GROUP) ||
	    (parsing_mode == CREATE_GROUP) ||
	    (parsing_mode == XCOMMIT)) {
		Emit_Stateless_PLinda_Function(parsing_mode, c, plinda);
		return 1;
	}
   }
   Emit_Sundry_PLinda_calls(parsing_mode, c, plinda);
   if (gid_name != NULL) {
	Out_PLinda(gid_name, plinda);	
	Out_PLinda(",",plinda);	
   }
   return 0;
}





/******
Function:	Obtain_ARGRDP_Assignment_and_Emit
Purpose:	This function is called upon realization of the 
		parsing mode being ARG_RDP.  The variable that gets
		the return value of argrdp() is obtained through parsing
		and emitted into the plinda output file.
******/
void
Obtain_ARGRDP_Assignment_and_Emit(CONTEXT * c, FILE * plinda)
{
	char test, var_to_assign[80];
	int i = 0;
	CON_reset(c);
	do {
		test = CON_getc(c);
		var_to_assign[i++] = test;
		if (test == '\0') {
			fprintf(stderr,"Error:: Line %d:",c->line_number);
			fprintf(stderr,"pl_argrdp() must be assigned to integer.\n");
		}
	} while (test != '=');
	CON_reset(c);
	var_to_assign[i++] = ' ';
	var_to_assign[i] = '\0';
	Out_PLinda(var_to_assign, plinda);
}



/******
Function:	Emit_Pl_Command
Purpose:	Emit target code for all the PLinda Commands.
		(ie. PLinda Commands:
			pl_in[gid](tuple);
			pl_out[gid](tuple);
			pl_rd[gid](tuple);
			pl_rdp[gid](tuple);
		)
******/
int
Emit_Pl_Command(int parsing_mode, CONTEXT * c, FILE * plinda, SymTable * symT)
{
	SymTable * Sym_Table;
	int tuple_length;

/**	1) Load the Symbol Table 	**/
	Sym_Table = Present_SymTable(symT);
/*** 	For debugging, print out the symbol table 
	fprintf(stderr,"-----------Analyze_PRE_Print_SymTable-----------\n");
	Print_SymTable(Sym_Table);
	fprintf(stderr,"-----------Analyze_POST_Print_SymTable-----------\n");
***/


/**	2) Emit the function corresponding to the parsing mode.  Exit if the
	parsing mode and tuple space deem no tuple state to be outputted. 
**/
	if (parsing_mode == ARG_RDP)
		Obtain_ARGRDP_Assignment_and_Emit(c, plinda);
	if (Immediate_Stateless_Emission(parsing_mode, Sym_Table, c, plinda))
		return 1;


/**	3) PLinda function call will have tuple_space arguments.
	   Reset current position of 'cursor' to the beginning of 
	   tuple-space. 
**/
	CON_Align_to_Char('(', c);


/*** Note:	This could maybe be a little bit more robust.  ***/
	count_args = 0;
	tuple_length = TupleLength(c);
	if (tuple_length == -1)
		return (0);

/**	5) Collect and Emit all Tuple Space arguments.	**/
	CON_Align_to_Char('(', c);
	Emit_Beginning_of_Tuple_Space_Info(tuple_length, plinda);
	if (Collect_PLinda_TupleSpace(c, plinda, Sym_Table) == 0)
		return (0);
	Emit_End_of_Tuple_Space_Info(c, plinda);

	return 1;
}



/******
Function:	TupleLength
Purpose:	Returns the number of tuple arguments used.
		Figures it out by counting commas.
Note:		If case of error, -1 returned.
******/
int
TupleLength(CONTEXT * c)
{
	char test;
	int left_paren_builder = 0;
	int args = 1;
	test = CON_getc(c);
	while (test != ')')
	{
		test = CON_getc(c);
		if (test == '(')
			left_paren_builder = 1;
		if (test == ')')
			left_paren_builder = 0;
		if (test == ',')
			if (left_paren_builder == 0)
				args++;
		if (test == '\n')
		{
			fprintf(stderr,"Error: Missing closing ')' on line %d.\n",c->line_number);
			CON_print(c);
			return (-1);
		}
	}
/**
fprintf(stderr,"TupleLength => %d\n",args);
**/
	if (args == 0) {
		fprintf(stderr,"Error:: Line %d:  Tuple Argument space zero.\n", c->line_number);
		return (-1);
	}
	if (args > 5) {
		fprintf(stderr,"Error:: Line %d:  PLinda/Fortran Version 2.6b has maximum tuple space of 5.\n", c->line_number);
		return (-1);
	}
	return (args);
}



/*******
Function:	Check_if_End_of_Line
Purpose:	Returns 1 if from the marker until the end of the line is
		all whitespace, 0 otherwise.
Called From:	Collect_PLinda_TupleSpace()
*******/
int
Check_if_End_of_Line(CONTEXT * c)
{
	int i = 0;
	int j;
	char temp;
	while (1)
	{
		temp = CON_getc(c);
/**
		fprintf(stderr,"Check_if_End_of_Line()::temp => %c\n", temp);
**/
		i++;
		if ((temp=='\n') || (temp=='\0'))
			return 1;
		/** If it's not a space, but a character of substance, 
			then backup and return 0 **/
		if (!((temp==' ') || (temp=='\t')))
		{
			for (j=0; j < i; j++)
				CON_backup(c);
			return 0;
		}
	}
}




/*******
Function:	Collect_PLinda_TupleSpace
Purpose:	Collects each token argument of tuple space and 
		calls function Emit_PLinda_based_upon_Type() to figure out what
		type the string collected in this function is.
*******/
int
Collect_PLinda_TupleSpace(CONTEXT * c, FILE * plinda, SymTable * symT)
{
	char temp[80];
	int i, j;
	int left_paren_builder = 0;
	int No_Tuple_Args = 1;
	for (i = 0; i < 80; i++)
		temp[i] = '\0';

/**
fprintf(stderr,"\n\nCollect_PLinda_TupleSpace:: line => \n");
CON_print(c);
**/
	while (1)
	{
		i = 0;
		temp[i] = CON_getc(c);
/**
fprintf(stderr,"Collect_PLinda_TupleSpace:: First char => %c\n",temp[i]);
**/
		if ((Check_if_End_of_Line(c)) && (No_Tuple_Args == 1)) {
			temp[++i] = '\0';
/**
			fprintf(stderr,"Collect_PLinda_TupleSpace:: string in Question(top) => %s\n", temp);
**/
			Emit_PLinda_based_upon_Type(temp, plinda, c, symT);
			return 1;
		}
		while (1)
		{
			if ( (temp[i] == ',') && (left_paren_builder == 0) )
				break;
			if (temp[i] == '(')
				left_paren_builder = 1;
			if (temp[i] == ')')
				left_paren_builder = 0;
			if ((White_Space(temp[i]) == 0) && (!End_of_Line(temp[i])))
				i++;
			temp[i] = CON_getc(c);
/**
fprintf(stderr,"Collect_PLinda_TupleSpace:: char. => %c\n", temp[i]);
**/
			if ( (temp[i] == ')') && (Check_if_End_of_Line(c)) )
			{
				temp[++i] = '\0';
/**
				fprintf(stderr,"Collect_PLinda_TupleSpace:: string in Question(bottom) => %s\n", temp);
**/
				Emit_PLinda_based_upon_Type(temp, plinda, c, symT);
				return 1;
			}
		}
		No_Tuple_Args = 0;
		count_args++;
		temp[i] = '\0';
/**
		fprintf(stderr,"Collect_PLinda_TupleSpace:: string in Question(bottom.2) => %s\n", temp);
**/
		Emit_PLinda_based_upon_Type(temp, plinda, c, symT);
		Out_PLinda(",",plinda);
		if (count_args % 2 == 0)
		{
			Out_PLinda("\n     +  ",plinda);
		}
		for (j=0; j <= i; j++)
			temp[j] = '\0';
	}
}




/*******
Function:	Legal_PLinda_Char
Purpose:	Returns 1 if the character passed in is a legal(possible)
		character in the PLinda tuple argument space.
Called From:	Emit_PLinda_based_upon_Type()
*******/
int 
Legal_PLinda_Char(char a, int left_paren_builder)
{
/**
   fprintf(stderr,"Legal_PLinda_Char:: Legal Char => %c\n",temp[i]);
**/
   if (left_paren_builder)
	return 1;
   if (Alpha(a) || Number(a))
	return 1;
   if (a == '(')
	return 1;
   if ((a == ')') && (left_paren_builder))
	return 1;
   if ( (a == ':') || (a == '*') )
	return 1;
   return 0;
}




/*******
Function:	Fetch_Type_and_Size
Purpose:	Look up the variable name in the symbol table and
		give values to call-by-reference parameters 'var_type'
		and 'var_size'.
*******/
void
Fetch_Type_and_Size(SymTable * symT, char var_name[80], 
			int * var_type, int * var_size, CONTEXT * c)
{
/* 1) Symbol Table lookup of the variable. */
   List * var_list;
/**
   fprintf(stderr,"Fetch_Type_and_Size: varname to look up => %s\n",var_name);
**/
   var_list = Access_List(symT->list, var_name);

/* 2) If var_name can't be found in the Symbol Table, then the variable
	is an Fortran implicit type(based on first letter of variable). */
   if (var_list == NULL)
   {
	*var_type = Retrieve_from_Implicit(symT, var_name[0]); 
	*var_size = 0;
/**
	fprintf(stderr,"Fetch_Type_and_Size: Implicit var. => %s ?\n",var_name);
	fprintf(stderr,"Fetch_Type_and_Size: Implicit var_type => %d\n",var_type);
**/
	if (*var_type == 0)
	{
		fprintf(stderr,"Fetch_Type_and_Size:: Error Line %d: '%s' - Undeclared variable.\n", c->line_number, var_name);
	}
   }
   else
   {
/* 2) Else the variable was found in the Symbol Table.
	Assign the size and type of the variable. */
	*var_type = var_list->variable_type; 
	*var_size = var_list->variable_size;
   }

/**
   fprintf(stderr,"Fetch_Type_and_Size:: var_type => %d\n",*var_type);
   fprintf(stderr,"Fetch_Type_and_Size:: var_size => %d\n\n",*var_type);
**/
}


/*******
Function:	Analyze_for_Type_and_Size_Emittance
Purpose:	Calls the function to emit the PLinda variable type function
		and its size according to its array size (or lack thereof).
*******/
void
Analyze_for_Type_and_Size_Emittance(char * array_size, int var_type, 
					char var_name[80], int l_value, 
					FILE * plinda, CONTEXT * c)
{
/* 1) Emit a PLinda function call specific to type. 
	A return of 1 means all tuple space info has been specially emitted */
   if (Emit_According_to_Var_Type(var_type, var_name, l_value, plinda, c))
	return;

   Out_PLinda(var_name, plinda);
   Out_PLinda(",",plinda);

   if (array_size == NULL)
	Out_PLinda("0", plinda);
   else {
/**
	fprintf(stderr,"Analyze_for_Type_Size_Emittance:: array_size => %s\n",array_size);
**/
	Out_PLinda(array_size, plinda);
   }
   Out_PLinda(")",plinda);
}





/*******
Function:	Get_Array_Length
Purpose:	Colon in tuple space signifies an array of specified length
*******/ 
char *
Get_Array_Length(char line[80], int index)
{
	char * array_length, * start_of_array;
	array_length = (char *)malloc(sizeof(char) * 80);
	start_of_array = array_length;
	while (line[++index] != '\0')
		*array_length++ = line[index];
	*array_length = '\0';
	return (start_of_array);
}


/*******
Function:	Fetch_lvalue_Spec	
Purpose:	Return 1 if the PLinda programmer has placed a '?' question mark
		in front of a variable denoting lvalue insertion desired.
		Return 0, otherwise.
Called By:	Emit_PLinda_based_upon_Type()
*******/ 
int
Fetch_lvalue_Spec(char tuple_space_str_arg[80])
{
	int i = 0;
	while (White_Space(tuple_space_str_arg[i]))
		i++;
	if (tuple_space_str_arg[i] == '?')
		return 1;
	else 
		return 0;
}



/*******
Function:	Fetch_Explicit_Array_Size
Purpose:	Return the string which specifies the array size specified
		by the PLinda programmer after the PLinda ':' size operator.
Called By:	Emit_PLinda_based_upon_Type()
*******/
char *
Fetch_Explicit_Array_Size(char tuple_space_str_arg[80])
{
   char * array_size, * Array_Size;
   int i = 0;
   array_size = (char *)malloc(sizeof(char) * 80);
   Array_Size = array_size;

/**
   fprintf(stderr,"\nFetch_Explicit_Array_Size:: tuple_space_str_arg => %s\n",tuple_space_str_arg);
**/

/** Align to the beginning of the array size specification **/
   while (tuple_space_str_arg[i++] != ':') {	
	if (tuple_space_str_arg[i] == '\0') {
		return (NULL);
	}
   }
	
/** Collect the array size **/
   while (1) {
	if (	(tuple_space_str_arg[i] == ')') ||
		(tuple_space_str_arg[i] == '\0')
						) {
		*array_size = '\0';
/**
		fprintf(stderr,"\tFetch_Explicit_Array_Size:: !Determined Array_Size => %s\n",Array_Size);
**/
		return (Array_Size);
	}
  	*array_size++ = tuple_space_str_arg[i++];
   }
}



/*******
Function:	Fetch_Pure_Var_Name
Purpose:	Returns the variable name (without any array arguments or references),
		so it can later be looked up in the Symbol Table.

Note:		3 cases of stopping collection of a variable name:
			(0)	'('   Array index specification
			(1)	')'   End of tuple space arguments  
			(2) 	':'   Array size specification
			(3) 	'\0'  End of line in the string
Called By:	Emit_PLinda_based_upon_Type()
*******/
char *
Fetch_Pure_Var_Name(char tuple_space_str_arg[80])
{
	char * var_name, * Var_Name;
	int i = 0;
	var_name = (char *)malloc(sizeof(char) * 80);
	Var_Name = var_name;
	while ((White_Space(tuple_space_str_arg[i])) || (tuple_space_str_arg[i] == '?'))
		i++;
	
	while (1) {
		if (	(tuple_space_str_arg[i] == '(') || 
			(tuple_space_str_arg[i] == ')') ||
			(tuple_space_str_arg[i] == ':') ||
			(tuple_space_str_arg[i] == '\0')
							) {
			*var_name = '\0';
			return (Var_Name);
		}
		*var_name++ = tuple_space_str_arg[i++];
	}
}


/*******
Function:	Fetch_Var_Name_and_Tuple_Space_Arg
Purpose:	Returns the full tuple space argument to used (variable name and
		any other info. like array indexes).
Note:		Variable 'left_paren_builder' is used as a counter to await a final
		terminating  right parentheses ')'.
Called By:	Emit_PLinda_based_upon_Type()
*******/
char *
Fetch_Var_Name_and_Tuple_Space_Arg(char tuple_space_str_arg[80])
{
   char * ts_arg , * TS_Arg;
   int i, left_paren_builder;
   i = 0;
   left_paren_builder = 1;
   ts_arg = (char *)malloc(sizeof(char) * 80);
   TS_Arg = ts_arg;
   while ((White_Space(tuple_space_str_arg[i])) || (tuple_space_str_arg[i] == '?'))
	i++;
	
   while (1) {
 	if (tuple_space_str_arg[i] == '(') 
		left_paren_builder++;
	if (tuple_space_str_arg[i] == ')') 
		left_paren_builder--;
	if (
		(tuple_space_str_arg[i] == '\0') 
						|| 
		(	(tuple_space_str_arg[i] == ',')  &&
			(left_paren_builder == 1)	)
						|| 
		(	(tuple_space_str_arg[i] == ':')  &&
			(left_paren_builder == 1)	)
						|| 
	   	(	(tuple_space_str_arg[i] == ')')  && 
			(left_paren_builder == 0)	)
							  ) {
		/** End of specified tuple_space arg. str. **/
		*ts_arg = '\0';
/**
		fprintf(stderr,"Fetch_Var_Name_and_Tuple: ret => %s\n",TS_Arg);
**/
		return (TS_Arg);
	}
	*ts_arg++ = tuple_space_str_arg[i++];
   }
}



/*******
Function:	Emit_PLinda_based_upon_Type
Purpose:	Takes the character string passed to it and takes action 
		based on the  following:
 		  (1)	Constant (Int, Real, Char. String, etc.)
		  (2)	Variable (Look up in the Symbol Table)
		Emits PLinda function call code for both (1), (2)
Note:		Need to clean this code up!!
*******/
int
Emit_PLinda_based_upon_Type(char tuple_space_str_arg[80], FILE * plinda, CONTEXT * c, SymTable * symT)
{
	char * var_argument_in_Full, * var_name, * array_length_specified;
	int var_type, var_size;
	var_type = var_size =  0;
	array_length_specified = NULL;

/**
fprintf(stderr,"\nEmit_PLinda_based_Upon_Type: tuple_space_str_arg => %s\n",tuple_space_str_arg );
**/

/**	(0) Get the full tuple space argument: variable name, array indices, etc. **/
	var_argument_in_Full = Fetch_Var_Name_and_Tuple_Space_Arg(tuple_space_str_arg);

/** 	(1) Check if constant in tuple-space. If so, emit code and exit. **/ 
	if (Constants_Emit(var_argument_in_Full, plinda))
		return 1;

/**	(2) Get explicit array length specification (if there is one) **/
	array_length_specified = Fetch_Explicit_Array_Size(tuple_space_str_arg);


/**	(3) Get the variable name for later symbol table lookup **/
	var_name = Fetch_Pure_Var_Name(tuple_space_str_arg);

	
/**
fprintf(stderr,"Analyze:: Pre-Fetch: var_name => %s\n",var_name);
fprintf(stderr,"Analyze:: Pre-Fetch: var_argument_in_Full => %s\n",var_argument_in_Full);
**/

/*** 	(4) Load variable type and variable size ***/
	Fetch_Type_and_Size(symT, var_name, &var_type, &var_size, c);


/*** 	(5) Emit PLinda variable type function and variable size(array size) ***/
	Analyze_for_Type_and_Size_Emittance(array_length_specified, var_type, 
			var_argument_in_Full, Fetch_lvalue_Spec(tuple_space_str_arg),
			 plinda, c);
	return 1;
}




/********
Function:	Constants_Emit
Purpose:	Checks if the char. array is a constant.  Emits relevant
		command if so.
Need to Fix/Do:	Errors outputted for missing <"hello>   end of quotation!!
			-not yet...
Need to Fix/Do II:	Need to find the size of the constant field entered
			and make it a double it is too large for a single
			field.
********/
int
Constants_Emit(char temp[80], FILE * plinda)
{
	int i = 0;
	int double_precise = 0;

/** Case 1.	Constant String **/
	if (temp[i] == '"') {
		Emit_Constant_String(temp, plinda);
		return 1;
	}


/** Case 2.	Constant Integer or Float (or Double??) **/
	if ( (Number(temp[i])) || (temp[i] == '.') || (temp[i] == '-') ) {
		while (temp[i] != '\0') {
			if ((temp[i] == 'D') || (temp[i] == 'd'))
				double_precise = 1;
			i++;
		}
		i = 0;
		while (temp[i] != '\0') {
			if (temp[i] == '.') {
				if (double_precise == 1)
					Emit_Constant_Double(temp,plinda);
				else
					Emit_Constant_Float(temp,plinda);
				return 1;
			}
			i++;
		}
		Emit_Constant_Int(temp, plinda);
		return 1;
	}
	return 0;
}

