/************
File:		List.c
Last Revision:	11/3/95
Purpose:	List for the Symbol Table
*************/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "List.h"
#include "DebugError.h"


/*******
Function:	Create_List
Purpose:	To malloc space for the List.
*******/
List *
Create_List()
{
	List * ret;
	ret = (List *) malloc(sizeof(List));
	if (ret==NULL)
	{	
		fprintf(stderr,"Compiler Error:: Create_List malloc failed\n");
		return (NULL);
	}
	ret->variable_type = -1;
	ret->variable_name = NULL;
	ret->variable_size = 0;
	ret->var_opt_len = 0;
	ret->range_a = 0;
	ret->range_b = 0;
	ret->next = NULL;
	ret->implicit_type = 1;	 
	return (ret);
}




/*******
Function:	Insert_List
Purpose:	To create a list node to insert the fields into the list.
*******/
List *
Insert_List(List * list, char * variable_name, int type, int size, int var_opt_len)
{
	List * tail;
	List * head = list;
	if (head == NULL)
	{
		head = Create_List();
		head->variable_type = type;
		head->variable_name = strdup(variable_name);
		head->variable_size = size;
		head->var_opt_len = var_opt_len;
	}	
	else
	{
		tail = Access_List(list, variable_name);
		if (tail != NULL)
			tail->variable_type = type;  	/** Common Block!! **/
		else
		{
			tail = head;
			while (tail->next != NULL)
				tail = tail->next;
			tail->next = Create_List();
			tail->next->variable_type = type;
			tail->next->variable_name = strdup(variable_name);
			tail->next->variable_size = size;
			tail->next->var_opt_len = var_opt_len;
		}
	}
	return (head);
}



/*******
Function:	Insert_List_Implicit
Purpose:	To create a list node to store the implicit stmt.
*******/
List *
Insert_List_Implicit(List * list, char range_a, char range_b, int type, int impl_type)
{
	List * tail;
	List * head = list;
	if (head == NULL)
	{
		head = Create_List();
		head->variable_type = type;
		head->variable_name = strdup("implicit");
		head->variable_size = 0;
		head->range_a = range_a;
		head->range_b = range_b;
		head->implicit_type = impl_type;	 
	}	
	else
	{
			tail = head;
			while (tail->next != NULL)
				tail = tail->next;
			tail->next = Create_List();
			tail->next->variable_type = type;
			tail->next->variable_name = strdup("implicit");
			tail->next->variable_size = 0;
			tail->next->range_a = range_a;
			tail->next->range_b = range_b;
			tail->next->implicit_type = impl_type;	 
	}
	return (head);
}



/*******
Function:	Access_List
Purpose:	To return the node containing the variable name matching
		that of the string passed in.  Otherwise, return NULL.
*******/
List *
Access_List(List * list, char * variable_name)
{
	List * temp = list;
	if (list==NULL)
		return (NULL);
	while (temp != NULL)
	{
		if (!strcmp(temp->variable_name,variable_name))
			return (temp);
		temp = temp->next;
	}
	return (NULL);
}



/******
Function:	Common_Block_Addition
Purpose:	The declaration of a common block is an unusual part of 
		Fortran.  A common block defines in a general manner
		variables which have no type at all, for they are required
		to be declared formally later.
Note:		With arrays, the array size (# of dimensions has) already
		been declared, so we don't redo this aspect of the var.
		declaration.  Only need it's type.
******/
void
Common_Block_Addition(List * list, char * variable_name, int type, int size)
{
	list->variable_type = type;  	
}




/******
Function:       Print_List
Purpose:        Debugging phase - print fields of the List
******/
void
Print_List(List * list)
{
        List * temp = list;
        Out("\n\t\tPrint list ::\n");
        while (temp != NULL)
        {
                Out("\t\t\tVariable Name::  ");
                Out(temp->variable_name);
		Out("\n");
                Out("\t\t\t\tVariable Array Size::  ");
                Out_Int(temp->variable_size);
                Out("\n");
                Out("\t\t\t\tVariable Optional Length::  ");
                Out_Int(temp->var_opt_len);
                Out("\n");
                Out("\t\t\t\tVariable Type::  ");
                Out_Int(temp->variable_type);
                Out("\n");
		if (temp->range_a != 0)
		{
                	Out("\t\t\tVariable Range::  ");
                	Out_Char(temp->range_a);
			if (temp->implicit_type)
                		Out("-");
			else
                		Out(",");
                	Out_Char(temp->range_b);
                	Out("\n");
		}
                temp = temp->next;
        }
        Out("\t\tEnd Print List.\n");
}

