/************
File:		SymTable.c
Last Revision:	11/3/95
Purpose:	Symbol Table List of Lists.

*************/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "SymTable.h"
#include "DebugError.h"
#include "List.h"
#include "Char_Match.h"
#include "y.tab.h"



/*******
Function:	Create_SymTable
Purpose:	To malloc space for the SymTable
*******/
SymTable *
Create_SymTable()
{
	SymTable * ret;
	ret = (SymTable *) malloc(sizeof(SymTable));
	if (Error(ret,"Create_SymTable", "ret"))
		return (NULL);
	ret->name = NULL;
	ret->list = NULL;
	ret->implicit = NULL;
	ret->next = NULL;
	return (ret);
}




/*******
Function:	Insert_SymTable
Purpose:	To create a SymTable node to insert the subroutine/common block name
		into the node, and create the potential for the list of variables
		in that subroutine/common block.
*******/
SymTable *
Insert_SymTable(SymTable * symT, char * name)
{
	SymTable * tail;
	if (symT == NULL)
	{
		tail = Create_SymTable();
		tail->name = strdup(name);
		return (tail);
	}	
	else
	{
		tail = symT;
		while (tail->next != NULL)
			tail = tail->next;
		tail->next = Create_SymTable();
		tail->next->name = strdup(name);
		return (symT);
	}
}



/*******
Function:	Access_SymTable
Purpose:	To return the node containing the variable name matching
		that of the string passed in.  Otherwise, return NULL.
*******/
SymTable *
Access_SymTable(SymTable * list, char * name)
{
	SymTable * temp = list;
	if (list==NULL)
		return (NULL);
	while (temp != NULL)
	{
		if (!strcmp(temp->name,name))
			return (temp);
		temp = temp->next;
	}
	return (NULL);
}



/*******
Function:	Last_SymTable
Purpose:	Return the last node on the Symbol Table List.
*******/
SymTable *
Last_SymTable(SymTable * symT)
{
	SymTable * temp = symT;
	if (symT == NULL)
		return (NULL);
	if (symT->name == NULL)
		return (symT);
	while (temp->next != NULL)
	{
		if (temp->next->name == NULL)
		{
			fprintf(stderr,"Compiler Error: Symbol Table Name NULL on symTable Entry\n"); 
			return (temp->next);
		}
		temp = temp->next;
	}
	return (temp);
}



/*******
Function:	Present_SymTable
Purpose:	Return the present Symbol Table being used (in context).
*******/
SymTable *
Present_SymTable(SymTable * symT)
{
	if (symT == NULL)
		return (NULL);
	if (symT->name == NULL)
	{
		fprintf(stderr,"Compiler Error:: Present Symbol Table Name is NULL.\n");
		return (NULL);
	}
	return (Last_SymTable(symT));
}



/******
Function:	Print_SymTable
Purpose:	Debugging purposes of printing the names on each node in the Symbol Table.
******/
void
Print_SymTable(SymTable * SymbolTable)
{
	SymTable * temp = SymbolTable;
	Out("\n\nPrint Symbol Table::\n");
	while (temp != NULL)
	{
		Out("  Subroutine/Common block Name:: ");
		Out("\t");
		Out(temp->name);
		Out("\n");
		Print_List(temp->list);
		Out("\tImplicit Stmts inside the Subroutine:");
		Print_List(temp->implicit);
		Out("\n");
		temp = temp->next;
	}
	Out("\n\nEnd Print Symbol Table::\n\n");
}




/*******
Function:	Retrieve_from_Implicit
Purpose:	Return the type (Integer defined in file <y.tab.h>) that corresponds
		with the Implicit defn.
Note:		If the implicit definition is not set up for that character, then
		the standard implicit type definitions of variables is followed
		as defined in Fortran 77.
*******/
int
Retrieve_from_Implicit(SymTable * SymT, char buff)
{
	List * i_list;
	if (SymT->implicit == NULL)
		return (Determine_Implicit_Type(buff));
	else
	{
		i_list= SymT->implicit;
		while (i_list != NULL)
		{
			if (i_list->implicit_type == COMMA)
			{
			   if ((buff == i_list->range_a) || (buff == i_list->range_b))
				return (i_list->variable_type);
			}
			else
			{
			   if ((buff >= i_list->range_a) && (buff <= i_list->range_b))
				return (i_list->variable_type);
			}
			i_list = i_list->next;
		}
	}
	return 0;
}


/*****
Function:	Determine_Implicit_Type
Purpose:	Based on the character passed to this function, the return
		type is determined by the 'implicit' declaration rule
		set forth by Fortran specifications.
Reference:	Structured Fortran 77, by Etter, page 22.
Ascii Character Set:	
*****/
int
Determine_Implicit_Type(char letter)
{
/*
fprintf(stderr,"\n\n\n\tletter => %c\n\n",letter);
*/

	if (!Alpha(letter))
	{	
		fprintf(stderr,"Error:  Variable, first letter has to be alphabetic character.\n");
		return 0;
	}

/** (I-N) OR (i-n) **/
	if (	((letter >= 73) && (letter <= 78)) ||
		((letter >= 105) && (letter <= 110))	)
		return (INTEGER);
/** (A-H) OR (a-h) OR (O-Z) OR (o-z) **/
	else
		return (REAL);
}

