/******** 
File:		flex.c
Authors: 	Tom Brown	(brow7141@cs.nyu.edu) 
	 	Peter Wyckoff
Purpose:  	Main Program for the Fortran/PLinda compiler.
Last Altered:	11.15.95 
********/
#include <stdio.h>
#include "context.h"  
#include "Decorate_Generate.h"  

int line_number = 1;

/*****
Function:	main
Purpose:	The driver function of the PLinda/Fortran translator 
		has 2 distinct phases.  
			1. Decorate Phase - Check if the Fortran source code is 
						syntactically correct.
			2. Generate Phase - Actively translate the PLinda/Fortran
						code to Fortran with C++ wrappers.
******/
int
main(int argc,char * argv[])
{
	CONTEXT * c;
	FILE * fp, * save_fp;

   	if (argc < 2) 
	{
     		fprintf(stderr,"*** ERROR usage - plf77 <filename>\n");
     		exit(1); 
	}
   	if ((fp=fopen(argv[1],"r"))==NULL)
   	{
		fprintf(stderr,
			"Error: Can't open specified file - %s\n",argv[1]);
		exit(1);
   	}

/***
	1.  Decorate the code - Comment out all the PLinda functions, and use
	the existing F77 compiler to compile the code.
***/
  	c =  CON_create(fp);
	if (Decorate(c,argv[1]) == 0)
	{
		fprintf(stderr,"Error: PLinda/Fortran Compilation Failed.\n");
		exit(1);
	}
	fclose(fp);

   	if ((save_fp=fopen(argv[1],"r"))==NULL)
   	{
		fprintf(stderr,
			"Error: Can't open specified file - %s\n",argv[1]);
		exit(1);
   	}


/**
fprintf(stdout,"main: pre-Generate");
fflush(stdout);
**/

/** 
	2.  Generate the PLinda code 
**/
  	c =  CON_create(save_fp);
	Generate(c,argv[1]);
	fclose(save_fp);

/* 
Return 0 for successful completion of preprocessing -Needed!!
*/
	return(0); 
}

