#include "plinda_lib.h"
#include <stdarg.h>

extern "C" {
#include "fortran-wrappers.h"
}
// The assumption of the datatype routines is that their
// second param will be the number of elements ---
// if it is 0, then the datatype is taken to be a scalar

// eg frint(x,0) indicates a plain integer
// frint(x,1) indicates a one element array of integer.


extern "C" void freal_main_(void);

int real_main(int ,char **,char **)
{
  freal_main_(); 
  return 1;
}


/***********
Begin New F77/PLinda Version
4/15/96
***********/

const int max(const int x, const int y)
{
  return x > y ? x : y ;
}

enum myType {t_rstr, t_rchar, t_lchar, t_rint, t_lint, 
	     t_rfloat, t_lfloat, t_rdouble, t_ldouble,
             t_rgid, t_lgid};

class Type {
private:
  myType type;	 	
  void * Data;
  int storageSize;
  int numElements; // == 0 means this is a scalar
  
public:
  Type(myType mt, int num_elems, void * data) {
    type = mt;
    numElements = num_elems;
    assert(numElements >= 0);
    storageSize = -1;
    Data = data;
    //    int temp ;
    switch (type) {
    case t_rstr:
      numElements = (::strlen((char *)Data) + 1);
    case t_rchar:
      Data = new char[max(numElements,1)];
      storageSize = numElements * sizeof(char);
      memcpy(Data,data,max(sizeof(char),storageSize));
      break;
    case t_rint:
      storageSize = numElements * sizeof(int);
      Data = new int[max(1,numElements)];
      memcpy(Data,data,max(sizeof(int),storageSize));
      break;
    case t_rfloat:
      storageSize = numElements * sizeof(float);
      Data = new float[max(1,numElements)];
      memcpy(Data,data,max(sizeof(float),storageSize));
      break;
    case t_rdouble:
      storageSize = numElements * sizeof(double);
      Data = new double[max(1,numElements)];
      memcpy(Data,data,max(sizeof(double),storageSize));
      break;
    case t_rgid:
      storageSize = numElements * sizeof(gid);
      Data = new gid[max(1,numElements)];
      memcpy(Data,data,max(sizeof(gid),storageSize));
      break;
    case t_lchar:
    case t_lint:
    case t_lfloat:
    case t_ldouble:
    case t_lgid:
      storageSize = 0;
      break;
    default:
      assert(0);
    }
    assert(storageSize >= 0);
  }

  void set(int num){
    switch(type) {
    case t_rstr:
      assert(numElements * sizeof(char) == (unsigned)storageSize);
      _PLtemp->setActual(num, TupleField::PLchar, (const char *)Data, 
			 numElements);
      break;
    case t_rchar:
      assert(numElements == 0 || numElements * sizeof(char) ==  (unsigned)storageSize);
      if(numElements == 0) 
	_PLtemp->setActual(num, TupleField::PLchar,*(char *)Data);
      else
	_PLtemp->setActual(num, TupleField::PLchar, (const char *)Data, 
			   numElements);
	break;
    case t_lchar:
      assert(storageSize == 0);
      if(numElements == 0) 
	_PLtemp->setFormal(num, TupleField::PLchar,(char *)Data);
      else
	_PLtemp->setFormal(num, TupleField::PLchar, (char *)Data,numElements);
      break;
    case t_rint:
      assert(numElements == 0 || numElements * sizeof(int) ==  (unsigned)storageSize);
      if (numElements == 0)
	_PLtemp->setActual(num, TupleField::PLint, *(int *) Data);
      else	
	_PLtemp->setActual(num, TupleField::PLint, (const int *) Data, 
			   numElements);
      break;
    case t_lint:
      assert(storageSize == 0);
      if (numElements == 0)
	  _PLtemp->setFormal(num, TupleField::PLint, (int *) Data);
	else
	  _PLtemp->setFormal(num, TupleField::PLint, (int *) Data, 
			     numElements);
      break;
    case t_rfloat:
      assert(numElements == 0 || numElements * sizeof(float) ==  (unsigned)storageSize);
      if (numElements == 0)
	_PLtemp->setActual(num, TupleField::PLfloat, *(float *) Data);
      else	
	_PLtemp->setActual(num, TupleField::PLfloat, (const float *) Data, 
			   numElements);
      break;
    case t_lfloat:
      assert(storageSize == 0);
      if (numElements == 0)
	  _PLtemp->setFormal(num, TupleField::PLfloat, (float *) Data);
	else
	  _PLtemp->setFormal(num, TupleField::PLfloat, (float *) Data, 
			     numElements);
      break;
    case t_rdouble:
      assert(numElements == 0 || numElements * sizeof(double) ==  (unsigned)storageSize);
      if (numElements == 0)
	_PLtemp->setActual(num, TupleField::PLdouble, *(double *) Data);
      else	
	_PLtemp->setActual(num, TupleField::PLdouble, (const double *) Data, 
			   numElements);
      break;
    case t_ldouble:
      assert(storageSize == 0);
      if (numElements == 0)
	  _PLtemp->setFormal(num, TupleField::PLdouble, (double *) Data);
	else
	  _PLtemp->setFormal(num, TupleField::PLdouble, (double *) Data, 
			     numElements);
      break;
    case t_rgid:
      assert(numElements == 0 || numElements * sizeof(gid) ==  (unsigned)storageSize);
      if (numElements == 0)
	_PLtemp->setActual(num, TupleField::PLgid, *(gid *) Data);
      else	
	_PLtemp->setActual(num, TupleField::PLgid, (const gid *) Data, 
			   numElements);
      break;
    case t_lgid:
      assert(storageSize == 0);
      if (numElements == 0)
	  _PLtemp->setFormal(num, TupleField::PLgid, (gid *) Data);
	else
	  _PLtemp->setFormal(num, TupleField::PLgid, (gid *) Data, 
			     numElements);
      break;
    default:
      cout << "Failed assertion going on " << __FILE__ << __LINE__ << flush;
    }
  } 	

  int storeSize(){    return storageSize;  } 
};


extern "C"
int *flchar(char * x,int num)
{
  return (int*) (new Type(t_lchar, num, x));
}

extern "C"
int *frchar(char * x,int num)
{
  return(int*) (new Type(t_rchar, num, x));
}

extern "C" 
int *frstr(char *x)
{
  return (int*)(new Type(t_rstr, 0, x));
}


extern "C" 
int *flint(int* x, int num)
{
  return (int*)(new Type(t_lint, num, x));
}

extern "C"
int *frint(int* x, int num )
{
  return (int*)new Type(t_rint, num, x);
}

extern "C"
int *flfloat(float * x, int num)
{
  return (int*)(new Type(t_lfloat, num, x));
}

extern "C"
int *frfloat(float * x, int num)
{
  return (int*)(new Type(t_rfloat, num, x));
}

extern "C"
int *frdouble(double * x, int num)
{
  return (int*)(new Type(t_rdouble, num, x));
}

extern "C"
int *fldouble(double * x, int num)
{
  return (int*)(new Type(t_ldouble, num, x));
}

extern "C"
int *frgid(int * x, int num)
{
  return (int*)(new Type(t_rgid, num, (gid*)x));
}

extern "C"
int *flgid(int * x, int num)
{
  return (int*)(new Type(t_lgid, num, (int*)x));
}

extern "C"
void fout(int *g,int* ,int line, char *file)
{
  assert(sizeof(gid)  == 2 * sizeof(int));
  char *mine = new char[strlen(file) + 1];
  strcpy(mine,file);
  gid *myG = (gid*)g;
  PL_out(*myG,_PLtemp,line,file);
}

extern "C"
void fin(int *g,int *,int line, char *file)
{
  assert(sizeof(gid)  == 2 * sizeof(int));
  char *mine = new char[strlen(file) + 1];
  strcpy(mine,file);
  gid *myG = (gid*)g;
  PL_in(*myG,_PLtemp,line,file);
}
extern "C"
int finp(int *g,int* ,int line, char *file)
{
  int ret;
  assert(sizeof(gid)  == 2 * sizeof(int));
  char *mine = new char[strlen(file) + 1];
  strcpy(mine,file);
  gid *myG = (gid*)g;
  ret = PL_inp(*myG,_PLtemp,line,file);
  if(ret == true)
    return 1;
  else
    return 0;
}


extern "C"
void frd(int *g,int* ,int line, char *file)
{
  assert(sizeof(gid)  == 2 * sizeof(int));
  char *mine = new char[strlen(file) + 1];
  strcpy(mine,file);
  gid *myG = (gid*)g;
  PL_rd(*myG,_PLtemp,line,file);
}


extern "C"
int frdp(int *g,int* ,int line, char *file)
{
  int ret;
  assert(sizeof(gid)  == 2 * sizeof(int));
  char *mine = new char[strlen(file) + 1];
  strcpy(mine,file);
  gid *myG = (gid*)g;
  ret = PL_rdp(*myG,_PLtemp,line,file);
  if(ret == true)
    return 1;
  else
    return 0;
}

extern "C"
int fargrdp(int* ,int line, char *file)
{
  assert(sizeof(gid)  == 2 * sizeof(int));
  char *mine = new char[strlen(file) + 1];
  strcpy(mine,file);
  int ret  = PL_arg_rdp(_PLtemp,line,file);
  if(ret == true)
    return 1;
  else
    return 0;
}


extern "C"
int fxrecover(int* ,int line, char *file)
{
  char *mine = new char[strlen(file) + 1];
  strcpy(mine,file);
  int ret  = PL_xrecover(_PLtemp,line,file);
  if(ret == true)
    return 1;
  else
    return 0;
}

extern "C" int *ftuple_(int * num, ...)
{
  int accumulate_size = 0;
  va_list ap;
  Type * elems[*num];
  va_start(ap, num);
  int i;
  for(i = 0 ; i < *num ; i++)      {
    elems[i] = *(Type**) va_arg(ap, void *);
    accumulate_size += elems[i]->storeSize();
  }
  _PLtemp = Tuple::create(*num,accumulate_size);
  for( i = 0 ; i < *num ; i++)  {
    elems[i]->set(i);	
    delete elems[i];
  }
  return (int*)_PLtemp;
}


extern "C"
void fxstart(int line, const char * file)
{
  char *mine = new char[strlen(file) + 1];
  strcpy(mine,file);
  PL_xstart(line,mine);
}

extern "C"
void fxcommit(int *,int line,const char *file)
{
  char *mine = new char[strlen(file) + 1];
  strcpy(mine,file);
  PL_xcommit(_PLtemp,line,file);
}


extern "C"
void fxcommit_nostate(int line,const char *file)
{
  char *mine = new char[strlen(file) + 1];
  strcpy(mine,file);
  PL_xcommit(NULL,line,file);
}


extern "C"
void fproc_eval(int* ,int line, const char *file)
{
  char *mine = new char[strlen(file) + 1];
  strcpy(mine,file);
  PL_proc_eval(_PLtemp,line,file);
}


extern "C"
void fcreate_group(int *g,char *desc, int line, char *file)
{
  assert(sizeof(gid) == 2 * sizeof(int));
  char *myDesc = new char[::strlen(desc) + 1];
  ::strcpy(myDesc,desc);
  char *myFile = new char[::strlen(file) + 1];
  ::strcpy(myFile,file);
  gid temp = PL_create_group(myDesc,line,myFile);
  *(gid*)g = temp;
}


extern "C" 
void fdestroy_group(int *gint, int line, char *file)
{
  assert(sizeof(gid) == 2 * sizeof(int));
  char *myFile = new char[::strlen(file) + 1];
  ::strcpy(myFile,file);
  gid *gptr = (gid*)gint;
  PL_destroy_group(*gptr,line,myFile);
}













