// This may look like C code, but it is really -*- C++ -*-
//
// File:     ClientProcess.h
// Created:  
// Author:   Karp Joo Jeong
// Mail:     jeong@cs.nyu.edu
//
#ifndef _CLIENT_PROCESS_H_
#define _CLIENT_PROCESS_H_


// ACE headers

// PLinda header files & forward declarations
#include "plinda_ids.h"
#include "ErrorType.h"
#include "plinda_requests.h"
#include "Process.h"
#include "TupleHandle.h"
//#include "Replay.h"
#include "CheckpointManager.h"
#include "Header.h"


class ServerCommLink;
class DaemonProcess;
class Transaction;

// GNU
#include "GNU_Interface.h"


// ------------------------------------------------------------
// Private log size limit: 
// up to one mega bytes, we keep private log in memory. 
// ------------------------------------------------------------

const int PL_MAX_IN_MEMORY_LOG_SZ = 5000000;




class ClientProcess : public Process {

public:
  void needLogForCheckpoint(bool v) { needLogForCheckpoint_ = v; }
void destroy_replaylist_before_last_commit();
  // --------------------------------------------------
  // CONSTRUCTOR and DESTRUCTOR
  // --------------------------------------------------
  void stat_end();
  ClientProcess(procid my_pid, procid parent_pid, const physid& arg,int interactive);
  virtual ~ClientProcess(void);

  void daemon_process(DaemonProcess* dproc);
  DaemonProcess* daemon_process(void);
  int interactive() const { return interactive_; }
  // --------------------------------------------------
  // used for debugging purposes ...
  // --------------------------------------------------
  void executable(const char*);
  const char* executable(void) const;
  const char* file(void) const;
  void file(const char* name);
  unsigned lineno(void) const;
  void lineno(unsigned) ;
  Header::FT_ModeType ft_degree() { return cur_ft_level; }
  void ft_degree(Header::FT_ModeType ft) { cur_ft_level = ft; }


  // --------------------------------------------------
  // Communication Link
  // --------------------------------------------------

  // overwritten virtual functions.
  virtual void handle_request(Header*,Tuple*); 
    void destroy_current_request(Header*,Tuple*);

  //  HeaderAndTuple current_request(void) const;  
  //  void current_request(HeaderAndTuple);


  // --------------------------------------------------
  // Scheduler
  // --------------------------------------------------

  // overwritten virtual functions.
  //  virtual int ready(void);
  int xcommit(Tuple *continuation= NULL, int loop = 0, double secs = 0);
  // request type
  //  ClientRequest request_type(void) const;


  // ---------------------------------------------------
  // Tuple Groups
  // ---------------------------------------------------

  void close_groups(void);

  // ---------------------------------------------------
  // Argument Tuple and Private Log
  // ---------------------------------------------------

  const physid& arg_id(void) const;

  //  CheckpointManager::ChkptDirType cur_log_dir(void) const;
  //  void cur_log_dir(CheckpointManager::ChkptDirType);

  //  CheckpointManager::ChkptDirType last_log_dir(void) const;
  //  void last_log_dir(CheckpointManager::ChkptDirType);

  // check if the log exists in the last checkpoint directory.
  int restore_private_log(void);
  int checkpoint_private_log(void);

  // load private log from/store to a file.
  //  int load_private_log(CheckpointManager::ChkptDirType);
  //  int store_private_log(CheckpointManager::ChkptDirType);

  virtual int resume(const TupleHandle&);
  
  // -------------------------------------------------
  // Forced Termination
  // -------------------------------------------------

  void kill(void);
  void terminate(void);

  // --------------------------------------------------
  // Failure Detection and Reaction 
  // --------------------------------------------------

  // overridden virtual functions.
  virtual int check_failure(void);
  virtual void declare_failure(void);
  virtual void handle_failure(void);

  virtual void declare_error(ErrorType);
  void handle_error(void);

  void cleanup_state(void);
  void remove_persistent_identity(void);

  int request_continuation();
  void destroy_log_and_replaylist();
  Pix x_blk_request;
  Tuple *x_blk_tuple;
  Header *x_blk_header;
  TupleGroup *x_blk_group;
  int consistent(void) const { return x_consistent; }
  bool needLogForModeChange_;
  bool needLogForMigration_;
  bool needLogForCheckpoint_;
  bool outstandingLogRequest_;

  // ---------------------------------------------------
  // Printing
  // ---------------------------------------------------

  virtual String print(void) const;

protected:

  // ---------------------------------------------------
  // Transactions
  // ---------------------------------------------------

  Transaction* trans(void) const;
  void trans(Transaction*);


  // ---------------------------------------------------
  // Tuple Groups
  // ---------------------------------------------------

  // list of the tuple groups that this process opens.
  TupleGroup* find_group(const gid&);


  // ---------------------------------------------------
  // Argument Tuple and Private Log
  // ---------------------------------------------------

  void arg_id(const physid&);  

  int send_reply(Header&,const Tuple * = NULL);


private:
  const int interactive_;
  bool firstTime;
  //  DLList<Tuple*> outList;
  // ---------------------------------------------------
  // what executable is this process running...?
  // ---------------------------------------------------
  char*  x_exec_name;

  // ---------------------------------------------------
  // Those who have created this process
  // ---------------------------------------------------

  procid x_parent_id;
  DaemonProcess* x_daemon_proc;

  // ---------------------------------------------------
  // Current Transaction
  // ---------------------------------------------------

  Transaction* x_trans;
  Transaction* x_temp_trans;  // when this process runs no transaction.

  // if there is a temporary transaction created, then commit or abort it.
  Transaction* start_temp_trans(void);
  void commit_temp_trans(void);
  void abort_temp_trans(void);


  bool waitingForLog;
  Header *waitingH;
  Tuple *waitingT;

  TupleHandle x_tuple_handle;

  // ---------------------------------------------------
  // Argument tuple and Private Log
  // ---------------------------------------------------

  physid x_arg_id;

  Tuple* x_private_log;
  bool x_consistent;

  Header::FT_ModeType cur_ft_level;
  DLList<Tuple *> replayList;
  int replayListCommittedLength ;
  // which directory 
  CheckpointManager::ChkptDirType x_cur_log_dir;
  CheckpointManager::ChkptDirType x_last_log_dir;

  // ---------------------------------------------------
  // opened groups
  // ---------------------------------------------------
  TupleGroupList x_group_list;

  // ---------------------------------------------------
  // store the last request for monitoring purposes ..
  // ---------------------------------------------------

  //---------------------------------------------------
  // the file name and line number of the last plinda
  // operation processed by the server for this client.
  // ---------------------------------------------------
  char*        x_file;
  unsigned     x_lineno;

};

#ifndef _OUTLINE_
#include "ClientProcess.iC"
#endif

#endif // _CLIENT_PROCESS_H_
