// This may look like C code, but it is really -*- C++ -*-
//
// File:     Scheduler.C
// Created:  
// Author:   Karp Joo Jeong
// Mail:     jeong@cs.nyu.edu
//

// *** GNU stuff ***
#include <iostream.h>
#include <SLList.h>
#include "C_Interface.h"

#include "Process.h"
#include "MonitorProcess.h"
#include "Scheduler.h"
#include "ProcessManager.h"
#include "ObjectSpace.h"
#include "CheckpointManager.h"
#include "TupleGroupManager.h"
#include "EventLog.h"



#ifdef _OUTLINE_
#define inline
#include "Scheduler.iC"
#endif // _OUTLINE_


Scheduler::Scheduler() {
  failure_detection_flag = 0;
  closedown_flag = 0;
  x_scheduling_mode = REGULAR_SCHEDULING;
}


// ----------------------------------------------------------------------
// this function is called when the communication layer
// detects process failure, so that the failure can be immediately
// handled.
// ----------------------------------------------------------------------
void
Scheduler::detect_failure(void) {
  failure_detection_flag = 1;
}


// Scheduler::main_loop is the main driver routine of the PLinda server.
// The tasks of the function are:
// (1) activates the communication manager to check if messages
//     have been arrived on connections;
// (2) let process objects handle requests received if any;
// (3) ask process objects to check if their processes are still alive,
//     when the failure detection mode is activated;
// (4) let process objects send replies to their processes.
int
Scheduler::event_loop() {

  while(1){
    if (ObjectSpace::state() == ObjectSpace::SHUTDOWN_SERVER) 
      break;
    if (ObjectSpace::state() == ObjectSpace::NEEDS_REBOOT) {
      cerr << "I know I need to reboot, am doing it now!\n";

      // throw out all current info
      ObjectSpace::chkpt_mgr.getReadyToReboot();
      ObjectSpace::proc_mgr.getReadyToReboot();
      ObjectSpace::group_mgr.getReadyToReboot();
      //      ObjectSpace::trans_mgr.getReadyToReboot(); should be empty!

      ObjectSpace::state(ObjectSpace::RUNNING);
      // now reload the latest correct checkpoint
      if(!ObjectSpace::recover()) {
	cout << "main restart from the scratch" << endl << flush;
      }
    }
    if(ObjectSpace::proc_mgr.num_processes() == 0 && 
       ObjectSpace::batch_flag == 1) {
      break;
    }
    
    ObjectSpace::proc_mgr.schedule();

    // ------------------------------------------------------------------
    // check on connections. 
    // ------------------------------------------------------------------
    Time_Value timeout(ObjectSpace::link_default_timeout);
    int num = ObjectSpace::comm_mgr.handle_events(timeout);
    if(num == 0) {
      // no request received; good time to flush the log buffers
      // or checkpoint tuple space.
    }

    // --------------------------------------------------------------------
    // check if each process is alive.
    // --------------------------------------------------------------------
    if(failure_detection_flag) {
      if(ObjectSpace::proc_mgr.handle_failure() > 0) {
	
      }
      failure_detection_flag = 0;
    }

  } // while
  return 1;
}







